from distutils.core import setup

from objpool.version import __version__

setup(
    name='objpool',
    version=__version__,
    author='Synnefo development team',
    author_email='synnefo-devel@googlegroups.com',
    packages=['objpool'],
    scripts=[],
    url='https://github.com/grnet/objpool',
    license='LICENSE.txt',
    description='An object pooling library',
    long_description=open('README.txt').read(),
    install_requires=[],
)



# Copyright 2011-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from time import sleep

from kamaki.clients.cyclades.rest_api import CycladesRestClient
from kamaki.clients import ClientError


class CycladesClient(CycladesRestClient):
    """Synnefo Cyclades Compute API client"""

    def create_server(
            self, name, flavor_id, image_id,
            metadata=None, personality=None):
        """Submit request to create a new server

        :param name: (str)

        :param flavor_id: integer id denoting a preset hardware configuration

        :param image_id: (str) id denoting the OS image to run on virt. server

        :param metadata: (dict) vm metadata updated by os/users image metadata

        :param personality: a list of (file path, file contents) tuples,
            describing files to be injected into virtual server upon creation

        :returns: a dict with the new virtual server details

        :raises ClientError: wraps request errors
        """
        image = self.get_image_details(image_id)
        metadata = metadata or dict()
        for key in ('os', 'users'):
            try:
                metadata[key] = image['metadata'][key]
            except KeyError:
                pass

        return super(CycladesClient, self).create_server(
            name, flavor_id, image_id,
            metadata=metadata, personality=personality)

    def start_server(self, server_id):
        """Submit a startup request

        :param server_id: integer (str or int)

        :returns: (dict) response headers
        """
        req = {'start': {}}
        r = self.servers_action_post(server_id, json_data=req, success=202)
        return r.headers

    def shutdown_server(self, server_id):
        """Submit a shutdown request

        :param server_id: integer (str or int)

        :returns: (dict) response headers
        """
        req = {'shutdown': {}}
        r = self.servers_action_post(server_id, json_data=req, success=202)
        return r.headers

    def get_server_console(self, server_id):
        """
        :param server_id: integer (str or int)

        :returns: (dict) info to set a VNC connection to virtual server
        """
        req = {'console': {'type': 'vnc'}}
        r = self.servers_action_post(server_id, json_data=req, success=200)
        return r.json['console']

    def get_firewall_profile(self, server_id):
        """
        :param server_id: integer (str or int)

        :returns: (str) ENABLED | DISABLED | PROTECTED

        :raises ClientError: 520 No Firewall Profile
        """
        r = self.get_server_details(server_id)
        try:
            return r['attachments'][0]['firewallProfile']
        except KeyError:
            raise ClientError(
                'No Firewall Profile',
                details='Server %s is missing a firewall profile' % server_id)

    def set_firewall_profile(self, server_id, profile):
        """Set the firewall profile for the public interface of a server

        :param server_id: integer (str or int)

        :param profile: (str) ENABLED | DISABLED | PROTECTED

        :returns: (dict) response headers
        """
        req = {'firewallProfile': {'profile': profile}}
        r = self.servers_action_post(server_id, json_data=req, success=202)
        return r.headers

    def list_server_nics(self, server_id):
        """
        :param server_id: integer (str or int)

        :returns: (dict) network interface connections
        """
        r = self.servers_ips_get(server_id)
        return r.json['attachments']

    def get_server_stats(self, server_id):
        """
        :param server_id: integer (str or int)

        :returns: (dict) auto-generated graphs of statistics (urls)
        """
        r = self.servers_stats_get(server_id)
        return r.json['stats']

    def list_networks(self, detail=False):
        """
        :param detail: (bool)

        :returns: (list) id,name if not detail else full info per network
        """
        detail = 'detail' if detail else ''
        r = self.networks_get(command=detail)
        return r.json['networks']

    def list_network_nics(self, network_id):
        """
        :param network_id: integer (str or int)

        :returns: (list)
        """
        r = self.networks_get(network_id=network_id)
        return r.json['network']['attachments']

    def create_network(
            self, name,
            cidr=None, gateway=None, type=None, dhcp=False):
        """
        :param name: (str)

        :param cidr: (str)

        :param geteway: (str)

        :param type: (str) if None, will use MAC_FILTERED as default
            Valid values: CUSTOM, IP_LESS_ROUTED, MAC_FILTERED, PHYSICAL_VLAN

        :param dhcp: (bool)

        :returns: (dict) network detailed info
        """
        net = dict(name=name)
        if cidr:
            net['cidr'] = cidr
        if gateway:
            net['gateway'] = gateway
        net['type'] = type or 'MAC_FILTERED'
        net['dhcp'] = True if dhcp else False
        req = dict(network=net)
        r = self.networks_post(json_data=req, success=202)
        return r.json['network']

    def get_network_details(self, network_id):
        """
        :param network_id: integer (str or int)

        :returns: (dict)
        """
        r = self.networks_get(network_id=network_id)
        return r.json['network']

    def update_network_name(self, network_id, new_name):
        """
        :param network_id: integer (str or int)

        :param new_name: (str)

        :returns: (dict) response headers
        """
        req = {'network': {'name': new_name}}
        r = self.networks_put(network_id=network_id, json_data=req)
        return r.headers

    def delete_network(self, network_id):
        """
        :param network_id: integer (str or int)

        :returns: (dict) response headers

        :raises ClientError: 421 Network in use
        """
        try:
            r = self.networks_delete(network_id)
            return r.headers
        except ClientError as err:
            if err.status == 421:
                err.details = [
                    'Network may be still connected to at least one server']
            raise

    def connect_server(self, server_id, network_id):
        """ Connect a server to a network

        :param server_id: integer (str or int)

        :param network_id: integer (str or int)

        :returns: (dict) response headers
        """
        req = {'add': {'serverRef': server_id}}
        r = self.networks_post(network_id, 'action', json_data=req)
        return r.headers

    def disconnect_server(self, server_id, nic_id):
        """
        :param server_id: integer (str or int)

        :param nic_id: (str)

        :returns: (int) the number of nics disconnected
        """
        vm_nets = self.list_server_nics(server_id)
        num_of_disconnections = 0
        for (nic_id, network_id) in [(
                net['id'],
                net['network_id']) for net in vm_nets if nic_id == net['id']]:
            req = {'remove': {'attachment': '%s' % nic_id}}
            self.networks_post(network_id, 'action', json_data=req)
            num_of_disconnections += 1
        return num_of_disconnections

    def disconnect_network_nics(self, netid):
        """
        :param netid: integer (str or int)
        """
        for nic in self.list_network_nics(netid):
            req = dict(remove=dict(attachment=nic))
            self.networks_post(netid, 'action', json_data=req)

    def _wait(
            self, item_id, current_status, get_status,
            delay=1, max_wait=100, wait_cb=None):
        """Wait for item while its status is current_status

        :param server_id: integer (str or int)

        :param current_status: (str)

        :param get_status: (method(self, item_id)) if called, returns
            (status, progress %) If no way to tell progress, return None

        :param delay: time interval between retries

        :param wait_cb: if set a progress bar is used to show progress

        :returns: (str) the new mode if successful, (bool) False if timed out
        """
        status, progress = get_status(self, item_id)

        if wait_cb:
            wait_gen = wait_cb(max_wait // delay)
            wait_gen.next()

        if status != current_status:
            if wait_cb:
                try:
                    wait_gen.next()
                except Exception:
                    pass
            return status
        old_wait = total_wait = 0

        while status == current_status and total_wait <= max_wait:
            if wait_cb:
                try:
                    for i in range(total_wait - old_wait):
                        wait_gen.next()
                except Exception:
                    break
            old_wait = total_wait
            total_wait = progress or total_wait + 1
            sleep(delay)
            status, progress = get_status(self, item_id)

        if total_wait < max_wait:
            if wait_cb:
                try:
                    for i in range(max_wait):
                        wait_gen.next()
                except:
                    pass
        return status if status != current_status else False

    def wait_server(
            self, server_id,
            current_status='BUILD',
            delay=1, max_wait=100, wait_cb=None):
        """Wait for server while its status is current_status

        :param server_id: integer (str or int)

        :param current_status: (str) BUILD|ACTIVE|STOPPED|DELETED|REBOOT

        :param delay: time interval between retries

        :max_wait: (int) timeout in secconds

        :param wait_cb: if set a progressbar is used to show progress

        :returns: (str) the new mode if succesfull, (bool) False if timed out
        """

        def get_status(self, server_id):
            r = self.get_server_details(server_id)
            return r['status'], (r.get('progress', None) if (
                            current_status in ('BUILD', )) else None)

        return self._wait(
            server_id, current_status, get_status, delay, max_wait, wait_cb)

    def wait_network(
            self, net_id,
            current_status='PENDING', delay=1, max_wait=100, wait_cb=None):
        """Wait for network while its status is current_status

        :param net_id: integer (str or int)

        :param current_status: (str) PENDING | ACTIVE | DELETED

        :param delay: time interval between retries

        :max_wait: (int) timeout in secconds

        :param wait_cb: if set a progressbar is used to show progress

        :returns: (str) the new mode if succesfull, (bool) False if timed out
        """

        def get_status(self, net_id):
            r = self.get_network_details(net_id)
            return r['status'], None

        return self._wait(
            net_id, current_status, get_status, delay, max_wait, wait_cb)

    def wait_firewall(
            self, server_id,
            current_status='DISABLED', delay=1, max_wait=100, wait_cb=None):
        """Wait while the public network firewall status is current_status

        :param server_id: integer (str or int)

        :param current_status: (str) DISABLED | ENABLED | PROTECTED

        :param delay: time interval between retries

        :max_wait: (int) timeout in secconds

        :param wait_cb: if set a progressbar is used to show progress

        :returns: (str) the new mode if succesfull, (bool) False if timed out
        """

        def get_status(self, server_id):
            return self.get_firewall_profile(server_id), None

        return self._wait(
            server_id, current_status, get_status, delay, max_wait, wait_cb)

    def get_floating_ip_pools(self):
        """
        :returns: (dict) {floating_ip_pools:[{name: ...}, ...]}
        """
        r = self.floating_ip_pools_get()
        return r.json

    def get_floating_ips(self):
        """
        :returns: (dict) {floating_ips: [fixed_ip: , id: , ip: , pool: ]}
        """
        r = self.floating_ips_get()
        return r.json

    def alloc_floating_ip(self, pool=None, address=None):
        """
        :param pool: (str) pool of ips to allocate from

        :param address: (str) ip address to request

        :returns: (dict) {
            fixed_ip: ..., id: ..., instance_id: ..., ip: ..., pool: ...}
        """
        json_data = dict()
        if pool:
            json_data['pool'] = pool
        if address:
            json_data['address'] = address
        r = self.floating_ips_post(json_data)
        return r.json['floating_ip']

    def get_floating_ip(self, fip_id):
        """
        :param fip_id: (str) floating ip id

        :returns: (dict)
            {fixed_ip: ..., id: ..., instance_id: ..., ip: ..., pool: ...},

        :raises AssertionError: if fip_id is emtpy
        """
        assert fip_id, 'floating ip id is needed for get_floating_ip'
        r = self.floating_ips_get(fip_id)
        return r.json['floating_ip']

    def delete_floating_ip(self, fip_id=None):
        """
        :param fip_id: (str) floating ip id (if None, all ips are deleted)

        :returns: (dict) request headers

        :raises AssertionError: if fip_id is emtpy
        """
        assert fip_id, 'floating ip id is needed for delete_floating_ip'
        r = self.floating_ips_delete(fip_id)
        return r.headers

    def attach_floating_ip(self, server_id, address):
        """Associate the address ip to server with server_id

        :param server_id: (int)

        :param address: (str) the ip address to assign to server (vm)

        :returns: (dict) request headers

        :raises ValueError: if server_id cannot be converted to int

        :raises ValueError: if server_id is not of a int-convertable type

        :raises AssertionError: if address is emtpy
        """
        server_id = int(server_id)
        assert address, 'address is needed for attach_floating_ip'
        req = dict(addFloatingIp=dict(address=address))
        r = self.servers_action_post(server_id, json_data=req)
        return r.headers

    def detach_floating_ip(self, server_id, address):
        """Disassociate an address ip from the server with server_id

        :param server_id: (int)

        :param address: (str) the ip address to assign to server (vm)

        :returns: (dict) request headers

        :raises ValueError: if server_id cannot be converted to int

        :raises ValueError: if server_id is not of a int-convertable type

        :raises AssertionError: if address is emtpy
        """
        server_id = int(server_id)
        assert address, 'address is needed for detach_floating_ip'
        req = dict(removeFloatingIp=dict(address=address))
        r = self.servers_action_post(server_id, json_data=req)
        return r.headers



# Copyright 2012-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from itertools import product

from kamaki.clients import livetest, ClientError
from kamaki.clients.astakos import AstakosClient


class Astakos(livetest.Generic):
    def setUp(self):
        self.cloud = 'cloud.%s' % self['testcloud']
        self.client = AstakosClient(
            self[self.cloud, 'url'], self[self.cloud, 'token'])
        print 'WHA?', self['astakos', 'details']
        with open(self['astakos', 'details']) as f:
            self._astakos_details = eval(f.read())

    def test_authenticate(self):
        self._test_0010_authenticate()

    def _test_0010_authenticate(self):
        r = self.client.authenticate()
        self.assert_dicts_are_equal(r, self._astakos_details)

    def test_get_services(self):
        self._test_0020_get_services()

    def _test_0020_get_services(self):
        for args in (tuple(), (self[self.cloud, 'token'],)):
            r = self.client.get_services(*args)
            services = self._astakos_details['access']['serviceCatalog']
            self.assertEqual(len(services), len(r))
            for i, service in enumerate(services):
                self.assert_dicts_are_equal(r[i], service)
        self.assertRaises(ClientError, self.client.get_services, 'wrong_token')

    def test_get_service_details(self):
        self._test_0020_get_service_details()

    def _test_0020_get_service_details(self):
        parsed_services = dict()
        for args in product(
                self._astakos_details['access']['serviceCatalog'],
                ([tuple(), (self[self.cloud, 'token'],)])):
            service = args[0]
            if service['type'] in parsed_services:
                continue
            r = self.client.get_service_details(service['type'], *args[1])
            self.assert_dicts_are_equal(r, service)
            parsed_services[service['type']] = True
        self.assertRaises(
            ClientError, self.client.get_service_details, 'wrong_token')

    def test_get_service_endpoints(self):
        self._test_0020_get_service_endpoints()

    def _test_0020_get_service_endpoints(self):
        parsed_services = dict()
        for args in product(
                self._astakos_details['access']['serviceCatalog'],
                ([], [self[self.cloud, 'token']])):
            service = args[0]
            if service['type'] in parsed_services:
                continue
            for endpoint, with_id in product(
                    service['endpoints'], (True, False)):
                vid = endpoint['versionId'] if (
                    with_id and endpoint['versionId']) else None
                end_args = [service['type'], vid] + args[1]
                r = self.client.get_service_endpoints(*end_args)
                self.assert_dicts_are_equal(r, endpoint)
            parsed_services[service['type']] = True
        self.assertRaises(
            ClientError, self.client.get_service_endpoints, 'wrong_token')

    def test_user_info(self):
        self._test_0020_user_info()

    def _test_0020_user_info(self):
        self.assertTrue(set([
            'roles',
            'name',
            'id']).issubset(self.client.user_info().keys()))

    def test_get(self):
        self._test_0020_get()

    def _test_0020_get(self):
        for term in ('id', 'name'):
            self.assertEqual(
                self.client.term(term, self[self.cloud, 'token']),
                self['astakos', term] or '')

    def test_list_users(self):
        self.client.authenticate()
        self._test_0020_list_users()

    def _test_0020_list_users(self):
        terms = set(['name', 'id'])
        uuid = 0
        for r in self.client.list_users():
            self.assertTrue(terms.issubset(r.keys()))
            self.assertTrue(uuid != r['id'] if uuid else True)
            uuid = r['id']



# Copyright 2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from mock import patch, call, MagicMock
from unittest import TestCase
from StringIO import StringIO
from datetime import datetime
#from itertools import product

from kamaki.cli import argument, errors
from kamaki.cli.config import Config


def assert_dicts_are_equal(test_case, d1, d2):
    for k, v in d1.items():
        test_case.assertTrue(k in d2)
        if isinstance(v, dict):
            test_case.assert_dicts_are_equal(v, d2[k])
        else:
            test_case.assertEqual(unicode(v), unicode(d2[k]))


cnf_path = 'kamaki.cli.config.Config'
arg_path = 'kamaki.cli.argument'


class Argument(TestCase):

    def test___init__(self):
        self.assertRaises(ValueError, argument.Argument, 'non-integer')
        self.assertRaises(AssertionError, argument.Argument, 1)
        self.assertRaises(AssertionError, argument.Argument, 0, 'noname')
        self.assertRaises(AssertionError, argument.Argument, 0, '--no name')
        self.assertRaises(AssertionError, argument.Argument, 0, ['-n', 'n m'])
        for arity, help, parsed_name, default in (
                (0, 'help 0', '--zero', None),
                (1, 'help 1', ['--one', '-o'], 'lala'),
                (-1, 'help -1', ['--help', '--or', '--more'], 0),
                (0, 'help 0 again', ['--again', ], True)):
            a = argument.Argument(arity, help, parsed_name, default)
            if arity:
                self.assertEqual(arity, a.arity)
            self.assertEqual(help, a.help)

            exp_name = parsed_name if (
                isinstance(parsed_name, list)) else [parsed_name, ]
            self.assertEqual(exp_name, a.parsed_name)

            exp_default = default if (default or arity) else False
            self.assertEqual(exp_default, a.default)

    def test_value(self):
        a = argument.Argument(1, parsed_name='--value')
        for value in (None, '', 0, 0.1, -12, [1, 'a', 2.8], (3, 'lala'), 'pi'):
            a.value = value
            self.assertEqual(value, a.value)

    def test_update_parser(self):
        for i, arity in enumerate((-1, 0, 1)):
            arp = argument.ArgumentParser()
            pname, aname = '--pname%s' % i, 'a_name_%s' % i
            a = argument.Argument(arity, 'args', pname, 42)
            a.update_parser(arp, aname)

            f = StringIO()
            arp.print_usage(file=f), f.seek(0)
            usage, exp = f.readline(), '[%s%s]\n' % (
                pname, (' %s' % aname.upper()) if arity else '')
            self.assertEqual(usage[-len(exp):], exp)
            del arp


class ConfigArgument(TestCase):

    def setUp(self):
        argument._config_arg = argument.ConfigArgument('Recovered Path')

    def test_value(self):
        c = argument._config_arg
        self.assertEqual(c.value, None)
        exp = '/some/random/path'
        c.value = exp
        self.assertTrue(isinstance(c.value, Config))
        self.assertEqual(c.file_path, exp)
        self.assertEqual(c.value.path, exp)

    def test_get(self):
        c = argument._config_arg
        c.value = None
        with patch('%s.get' % cnf_path, return_value='config') as get:
            self.assertEqual(c.value.get('global', 'config_cli'), 'config')
            self.assertEqual(get.mock_calls[-1], call('global', 'config_cli'))

    @patch('%s.keys' % cnf_path, return_value=(
        'image_cli', 'config_cli', 'history_cli', 'file'))
    def test_groups(self, keys):
        c = argument._config_arg
        c.value = None
        cset = set(c.groups)
        self.assertTrue(cset.issuperset(['image', 'config', 'history']))
        self.assertEqual(keys.mock_calls[-1], call('global'))
        self.assertFalse('file' in cset)
        self.assertEqual(keys.mock_calls[-1], call('global'))

    @patch('%s.items' % cnf_path, return_value=(
        ('image_cli', 'image'), ('file', 'pithos'),
        ('config_cli', 'config'), ('history_cli', 'history')))
    def test_cli_specs(self, items):
        c = argument._config_arg
        c.value = None
        cset = set(c.cli_specs)
        self.assertTrue(cset.issuperset([
            ('image', 'image'), ('config', 'config'), ('history', 'history')]))
        self.assertEqual(items.mock_calls[-1], call('global'))
        self.assertFalse(cset.issuperset([('file', 'pithos'), ]))
        self.assertEqual(items.mock_calls[-1], call('global'))

    def test_get_global(self):
        c = argument._config_arg
        c.value = None
        for k, v in (
                ('config_cli', 'config'),
                ('image_cli', 'image'),
                ('history_cli', 'history')):
            with patch('%s.get' % cnf_path, return_value=v) as gg:
                self.assertEqual(c.get_global(k), v)
                self.assertEqual(gg.mock_calls[-1], call('global', k))

    def test_get_cloud(self):
        c = argument._config_arg
        c.value = None
        with patch(
                '%s.get_cloud' % cnf_path,
                return_value='http://cloud') as get_cloud:
            self.assertTrue(len(c.get_cloud('mycloud', 'url')) > 0)
            self.assertEqual(get_cloud.mock_calls[-1],  call('mycloud', 'url'))
        with patch(
                '%s.get_cloud' % cnf_path,
                side_effect=KeyError('no token')) as get_cloud:
            self.assertRaises(KeyError, c.get_cloud, 'mycloud', 'token')
        invalidcloud = 'PLEASE_DO_NOT_EVER_NAME_YOUR_CLOUD_LIKE_THIS111'
        self.assertRaises(KeyError, c.get_cloud, invalidcloud, 'url')


class RuntimeConfigArgument(TestCase):

    def setUp(self):
        argument._config_arg = argument.ConfigArgument('Recovered Path')

    @patch('%s.Argument.__init__' % arg_path)
    def test___init__(self, arg):
        config, help, pname, default = 'config', 'help', 'pname', 'default'
        rca = argument.RuntimeConfigArgument(config, help, pname, default)
        self.assertTrue(isinstance(rca, argument.RuntimeConfigArgument))
        self.assertEqual(rca._config_arg, config)
        self.assertEqual(arg.mock_calls[-1], call(1, help, pname, default))

    @patch('%s.override' % cnf_path)
    def test_value(self, override):
        config, help, pname, default = argument._config_arg, 'help', '-n', 'df'
        config.value = None
        rca = argument.RuntimeConfigArgument(config, help, pname, default)
        self.assertEqual(rca.value, default)

        for options in ('grp', 'grp.opt', 'k v', '=nokey', 2.8, None, 42, ''):
            self.assertRaises(TypeError, rca.value, options)

        for options in ('key=val', 'grp.key=val', 'dotted.opt.key=val'):
            rca.value = options
            option, sep, val = options.partition('=')
            grp, sep, key = option.partition('.')
            grp, key = (grp, key) if key else ('global', grp)
            self.assertEqual(override.mock_calls[-1], call(grp, key, val))


class FlagArgument(TestCase):

    @patch('%s.Argument.__init__' % arg_path)
    def test___init__(self, arg):
        help, pname, default = 'help', 'pname', 'default'
        fa = argument.FlagArgument(help, pname, default)
        self.assertTrue(isinstance(fa, argument.FlagArgument))
        arg.assert_called_once(0, help, pname, default)


class ValueArgument(TestCase):

    @patch('%s.Argument.__init__' % arg_path)
    def test___init__(self, arg):
        help, pname, default = 'help', 'pname', 'default'
        fa = argument.ValueArgument(help, pname, default)
        self.assertTrue(isinstance(fa, argument.ValueArgument))
        arg.assert_called_once(1, help, pname, default)


class IntArgument(TestCase):

    def test_value(self):
        ia = argument.IntArgument(parsed_name='--ia')
        self.assertEqual(ia.value, None)
        for v in (1, 0, -1):
            ia.value = v
            self.assertEqual(ia.value, v)
        for v in ('1', '-1'):
            ia.value = v
            self.assertEqual(ia.value, int(v))
        for v, err in (
                ('invalid', errors.CLIError),
                (2.8, errors.CLIError),
                (923455555555555555555555555555555, errors.CLIError),
                (None, TypeError), (False, TypeError), ([1, 2, 3], TypeError)):
            try:
                ia.value = v
            except Exception as e:
                self.assertTrue(isinstance(e, err))


class DateArgument(TestCase):

    def test_timestamp(self):
        da = argument.DateArgument(parsed_name='--date')
        self.assertEqual(da.timestamp, None)
        date, format, exp = '24-10-1917', '%d-%m-%Y', -1646964000.0
        da._value = argument.dtm.strptime(date, format)
        self.assertEqual(da.timestamp, exp)

    def test_formated(self):
        da = argument.DateArgument(parsed_name='--date')
        self.assertEqual(da.formated, None)
        date, format, exp = (
            '24-10-1917', '%d-%m-%Y', 'Wed Oct 24 00:00:00 1917')
        da._value = argument.dtm.strptime(date, format)
        self.assertEqual(da.formated, exp)

    @patch('%s.DateArgument.timestamp' % arg_path)
    @patch('%s.DateArgument.format_date' % arg_path)
    def test_value(self, format_date, timestamp):
        da = argument.DateArgument(parsed_name='--date')
        self.assertTrue(isinstance(da.value, MagicMock))
        da.value = 'Something'
        format_date.assert_called_once(call('Something'))

    def test_format_date(self):
        da = argument.DateArgument(parsed_name='--date')
        for datestr, exp in (
                ('Wed Oct 24 01:02:03 1917', datetime(1917, 10, 24, 1, 2, 3)),
                ('24-10-1917', datetime(1917, 10, 24)),
                ('01:02:03 24-10-1917', datetime(1917, 10, 24, 1, 2, 3))):
            self.assertEqual(da.format_date(datestr), exp)
        for datestr, err in (
                ('32-40-20134', errors.CLIError),
                ('Wednesday, 24 Oct 2017', errors.CLIError),
                (None, TypeError), (0.8, TypeError)):
            self.assertRaises(err, da.format_date, datestr)


class VersionArgument(TestCase):

    def test_value(self):
        va = argument.VersionArgument(parsed_name='--version')
        self.assertTrue(va, argument.VersionArgument)
        va.value = 'some value'
        self.assertEqual(va.value, 'some value')


class RepeatableArgument(TestCase):

    @patch('%s.Argument.__init__' % arg_path)
    def test___init__(self, init):
        help, pname, default = 'help', 'pname', 'default'
        kva = argument.RepeatableArgument(help, pname, default)
        self.assertTrue(isinstance(kva, argument.RepeatableArgument))
        self.assertEqual(init.mock_calls[-1], call(-1, help, pname, default))


class KeyValueArgument(TestCase):

    @patch('%s.Argument.__init__' % arg_path)
    def test___init__(self, init):
        help, pname, default = 'help', 'pname', 'default'
        kva = argument.KeyValueArgument(help, pname, default)
        self.assertTrue(isinstance(kva, argument.KeyValueArgument))
        self.assertEqual(init.mock_calls[-1], call(-1, help, pname, default))

    def test_value(self):
        kva = argument.KeyValueArgument(parsed_name='--keyval')
        self.assertEqual(kva.value, [])
        for kvpairs in (
                'strval', 'key=val', 2.8, 42, None,
                ('key', 'val'), ('key val'), ['=val', 'key=val'],
                ['key1=val1', 'key2 val2'], ('key1 = val1', )):
            try:
                kva.value = kvpairs
            except Exception as e:
                self.assertTrue(isinstance(e, errors.CLIError))
        old = dict()
        kva = argument.KeyValueArgument(parsed_name='--keyval')
        for kvpairs, exp in (
                (('key=val', ), {'key': 'val'}),
                (['key1=val1', 'key2=val2'], {'key1': 'val1', 'key2': 'val2'}),
                (
                    ('k1=v1 v2', 'k3=', 'k 4=v4'),
                    {'k1': 'v1 v2', 'k3': '', 'k 4': 'v4'}),
                (('k=v1', 'k=v2', 'k=v3'), {'k': 'v3'})
            ):
            kva.value = kvpairs
            old.update(exp)
            assert_dicts_are_equal(self, kva.value, old)


class ProgressBarArgument(TestCase):

    class PseudoBar(object):
            message = ''
            suffix = ''

            def start():
                pass

    @patch('%s.FlagArgument.__init__' % arg_path)
    def test___init__(self, init):
        help, pname, default = 'help', '--progress', 'default'
        pba = argument.ProgressBarArgument(help, pname, default)
        self.assertTrue(isinstance(pba, argument.ProgressBarArgument))
        self.assertEqual(pba.suffix, '%(percent)d%%')
        init.assert_called_once(help, pname, default)

    def test_clone(self):
        pba = argument.ProgressBarArgument(parsed_name='--progress')
        pba.value = None
        pba_clone = pba.clone()
        self.assertTrue(isinstance(pba, argument.ProgressBarArgument))
        self.assertTrue(isinstance(pba_clone, argument.ProgressBarArgument))
        self.assertNotEqual(pba, pba_clone)
        self.assertEqual(pba.parsed_name, pba_clone.parsed_name)

    def test_get_generator(self):
        pba = argument.ProgressBarArgument(parsed_name='--progress')
        pba.value = None
        msg, msg_len = 'message', 40
        with patch('%s.KamakiProgressBar.start' % arg_path) as start:
            try:
                pba.get_generator(msg, msg_len)
                self.assertTrue(
                    isinstance(pba.bar, argument.KamakiProgressBar))
                self.assertNotEqual(pba.bar.message, msg)
                self.assertEqual(pba.bar.message, '%s%s' % (
                    msg, ' ' * (msg_len - len(msg))))
                self.assertEqual(pba.bar.suffix, '%(percent)d%% - %(eta)ds')
                start.assert_called_once()

                pba.get_generator(msg, msg_len, countdown=True)
                self.assertTrue(
                    isinstance(pba.bar, argument.KamakiProgressBar))
                self.assertNotEqual(pba.bar.message, msg)
                self.assertEqual(pba.bar.message, '%s%s' % (
                    msg, ' ' * (msg_len - len(msg))))
                self.assertEqual(pba.bar.suffix, '%(remaining)ds to timeout')
            finally:
                try:
                    pba.finish()
                except Exception:
                    pass

    def test_finish(self):
        pba = argument.ProgressBarArgument(parsed_name='--progress')
        pba.value = None
        self.assertEqual(pba.finish(), None)
        pba.bar = argument.KamakiProgressBar()
        with patch('%s.KamakiProgressBar.finish' % arg_path) as finish:
            pba.finish()
            finish.assert_called_once()


class ArgumentParseManager(TestCase):

    @patch('%s.ArgumentParseManager.parse' % arg_path)
    @patch('%s.ArgumentParseManager.update_parser' % arg_path)
    def test___init__(self, parse, update_parser):
        for arguments in (None, {'k1': 'v1', 'k2': 'v2'}):
            apm = argument.ArgumentParseManager('exe', arguments)
            self.assertTrue(isinstance(apm, argument.ArgumentParseManager))

            self.assertTrue(isinstance(apm.parser, argument.ArgumentParser))
            self.assertFalse(apm.parser.add_help)
            self.assertEqual(
                apm.parser.formatter_class,
                argument.RawDescriptionHelpFormatter)

            self.assertEqual(
                apm.syntax, 'exe <cmd_group> [<cmd_subbroup> ...] <cmd>')
            assert_dicts_are_equal(
                self, apm.arguments,
                arguments or argument._arguments)
            self.assertFalse(apm._parser_modified)
            self.assertEqual(apm._parsed, None)
            self.assertEqual(apm._unparsed, None)
            self.assertEqual(parse.mock_calls[-1], call())
            if arguments:
                update_parser.assert_called_once()

    def test_syntax(self):
        apm = argument.ArgumentParseManager('exe')
        self.assertEqual(
            apm.syntax, 'exe <cmd_group> [<cmd_subbroup> ...] <cmd>')
        apm.syntax = 'some syntax'
        self.assertEqual(apm.syntax, 'some syntax')

    @patch('%s.ArgumentParseManager.update_parser' % arg_path)
    def test_arguments(self, update_parser):
        apm = argument.ArgumentParseManager('exe')
        assert_dicts_are_equal(self, apm.arguments, argument._arguments)
        update_parser.assert_called_once()
        exp = {'k1': 'v1', 'k2': 'v2'}
        apm.arguments = exp
        assert_dicts_are_equal(self, apm.arguments, exp)
        self.assertEqual(update_parser.mock_calls[-1], call())
        try:
            apm.arguments = None
        except Exception as e:
            self.assertTrue(isinstance(e, AssertionError))

    @patch('%s.ArgumentParseManager.parse' % arg_path)
    def test_parsed(self, parse):
        apm = argument.ArgumentParseManager('exe')
        self.assertEqual(apm.parsed, None)
        exp = 'you have been parsed'
        apm._parsed = exp
        self.assertEqual(apm.parsed, exp)
        apm._parser_modified = True
        apm._parsed = exp + ' v2'
        self.assertEqual(apm.parsed, exp + ' v2')
        self.assertEqual(parse.mock_calls, [call(), call()])

    @patch('%s.ArgumentParseManager.parse' % arg_path)
    def test_unparsed(self, parse):
        apm = argument.ArgumentParseManager('exe')
        self.assertEqual(apm.unparsed, None)
        exp = 'you have been unparsed'
        apm._unparsed = exp
        self.assertEqual(apm.unparsed, exp)
        apm._parser_modified = True
        apm._unparsed = exp + ' v2'
        self.assertEqual(apm.unparsed, exp + ' v2')
        self.assertEqual(parse.mock_calls, [call(), call()])

    @patch('%s.Argument.update_parser' % arg_path
        )
    def test_update_parser(self, update_parser):
        apm = argument.ArgumentParseManager('exe')
        body_count = len(update_parser.mock_calls)
        exp = len(argument._arguments)
        self.assertEqual(body_count, exp)
        apm.update_parser()
        exp = len(apm.arguments) + body_count
        body_count = len(update_parser.mock_calls)
        self.assertEqual(body_count, exp)
        expd = dict(
            k1=argument.Argument(0, parsed_name='-a'),
            k2=argument.Argument(0, parsed_name='-b'))
        apm.update_parser(expd)
        body_count = len(update_parser.mock_calls)
        self.assertEqual(body_count, exp + 2)

    def test_update_arguments(self):
        (inp, cor, exp) = (
            {'k1': 'v1', 'k2': 'v3'}, {'k2': 'v2'}, {'k1': 'v1', 'k2': 'v2'})
        apm = argument.ArgumentParseManager('exe')
        with patch(
                '%s.ArgumentParseManager.update_parser' % arg_path) as UP:
            apm.update_arguments(None)
            self.assertEqual(len(UP.mock_calls), 0)
            apm._arguments = inp
            apm.update_arguments(cor)
            assert_dicts_are_equal(self, apm.arguments, exp)
            UP.assert_called_once_with()

    def test_parse(self):
        apm = argument.ArgumentParseManager('exe')
        parsed, unparsed = apm.parser.parse_known_args()
        apm.parse()
        self.assertEqual(apm._parsed, parsed)
        self.assertEqual(apm.unparsed, unparsed)


if __name__ == '__main__':
    from sys import argv
    from kamaki.cli.test import runTestCase
    runTestCase(Argument, 'Argument', argv[1:])
    runTestCase(ConfigArgument, 'ConfigArgument', argv[1:])
    runTestCase(RuntimeConfigArgument, 'RuntimeConfigArgument', argv[1:])
    runTestCase(FlagArgument, 'FlagArgument', argv[1:])
    runTestCase(FlagArgument, 'ValueArgument', argv[1:])
    runTestCase(IntArgument, 'IntArgument', argv[1:])
    runTestCase(DateArgument, 'DateArgument', argv[1:])
    runTestCase(VersionArgument, 'VersionArgument', argv[1:])
    runTestCase(RepeatableArgument, 'RepeatableArgument', argv[1:])
    runTestCase(KeyValueArgument, 'KeyValueArgument', argv[1:])
    runTestCase(ProgressBarArgument, 'ProgressBarArgument', argv[1:])
    runTestCase(ArgumentParseManager, 'ArgumentParseManager', argv[1:])



# Copyright 2011-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.

from base64 import b64encode
from os.path import exists
from io import StringIO
from pydoc import pager

from kamaki.cli import command
from kamaki.cli.command_tree import CommandTree
from kamaki.cli.utils import remove_from_items, filter_dicts_by_dict
from kamaki.cli.errors import (
    raiseCLIError, CLISyntaxError, CLIBaseUrlError, CLIInvalidArgument)
from kamaki.clients.cyclades import CycladesClient, ClientError
from kamaki.cli.argument import FlagArgument, ValueArgument, KeyValueArgument
from kamaki.cli.argument import ProgressBarArgument, DateArgument, IntArgument
from kamaki.cli.commands import _command_init, errors, addLogSettings
from kamaki.cli.commands import (
    _optional_output_cmd, _optional_json, _name_filter, _id_filter)


server_cmds = CommandTree('server', 'Cyclades/Compute API server commands')
flavor_cmds = CommandTree('flavor', 'Cyclades/Compute API flavor commands')
network_cmds = CommandTree('network', 'Cyclades/Compute API network commands')
ip_cmds = CommandTree('ip', 'Cyclades/Compute API floating ip commands')
_commands = [server_cmds, flavor_cmds, network_cmds, ip_cmds]


about_authentication = '\nUser Authentication:\
    \n* to check authentication: /user authenticate\
    \n* to set authentication token: /config set cloud.<cloud>.token <token>'

howto_personality = [
    'Defines a file to be injected to virtual servers file system.',
    'syntax:  PATH,[SERVER_PATH,[OWNER,[GROUP,[MODE]]]]',
    '  [local-path=]PATH: local file to be injected (relative or absolute)',
    '  [server-path=]SERVER_PATH: destination location inside server Image',
    '  [owner=]OWNER: virtual servers user id for the remote file',
    '  [group=]GROUP: virtual servers group id or name for the remote file',
    '  [mode=]MODE: permission in octal (e.g., 0777 or o+rwx)',
    'e.g., -p /tmp/my.file,owner=root,mode=0777']


class _service_wait(object):

    wait_arguments = dict(
        progress_bar=ProgressBarArgument(
            'do not show progress bar', ('-N', '--no-progress-bar'), False)
    )

    def _wait(
            self, service, service_id, status_method, current_status,
            countdown=True, timeout=60):
        (progress_bar, wait_cb) = self._safe_progress_bar(
            '%s %s: status is still %s' % (
                service, service_id, current_status),
            countdown=countdown, timeout=timeout)

        try:
            new_mode = status_method(
                service_id, current_status, max_wait=timeout, wait_cb=wait_cb)
            if new_mode:
                self.error('%s %s: status is now %s' % (
                    service, service_id, new_mode))
            else:
                self.error('%s %s: status is still %s' % (
                    service, service_id, current_status))
        except KeyboardInterrupt:
            self.error('\n- canceled')
        finally:
            self._safe_progress_bar_finish(progress_bar)


class _server_wait(_service_wait):

    def _wait(self, server_id, current_status, timeout=60):
        super(_server_wait, self)._wait(
            'Server', server_id, self.client.wait_server, current_status,
            countdown=(current_status not in ('BUILD', )),
            timeout=timeout if current_status not in ('BUILD', ) else 100)


class _network_wait(_service_wait):

    def _wait(self, net_id, current_status, timeout=60):
        super(_network_wait, self)._wait(
            'Network', net_id, self.client.wait_network, current_status,
            timeout=timeout)


class _firewall_wait(_service_wait):

    def _wait(self, server_id, current_status, timeout=60):
        super(_firewall_wait, self)._wait(
            'Firewall of server',
            server_id, self.client.wait_firewall, current_status,
            timeout=timeout)


class _init_cyclades(_command_init):
    @errors.generic.all
    @addLogSettings
    def _run(self, service='compute'):
        if getattr(self, 'cloud', None):
            base_url = self._custom_url(service) or self._custom_url(
                'cyclades')
            if base_url:
                token = self._custom_token(service) or self._custom_token(
                    'cyclades') or self.config.get_cloud('token')
                self.client = CycladesClient(base_url=base_url, token=token)
                return
        else:
            self.cloud = 'default'
        if getattr(self, 'auth_base', False):
            cyclades_endpoints = self.auth_base.get_service_endpoints(
                self._custom_type('cyclades') or 'compute',
                self._custom_version('cyclades') or '')
            base_url = cyclades_endpoints['publicURL']
            token = self.auth_base.token
            self.client = CycladesClient(base_url=base_url, token=token)
        else:
            raise CLIBaseUrlError(service='cyclades')

    def main(self):
        self._run()


@command(server_cmds)
class server_list(_init_cyclades, _optional_json, _name_filter, _id_filter):
    """List virtual servers accessible by user
    Use filtering arguments (e.g., --name-like) to manage long server lists
    """

    PERMANENTS = ('id', 'name')

    arguments = dict(
        detail=FlagArgument('show detailed output', ('-l', '--details')),
        since=DateArgument(
            'show only items since date (\' d/m/Y H:M:S \')',
            '--since'),
        limit=IntArgument(
            'limit number of listed virtual servers', ('-n', '--number')),
        more=FlagArgument(
            'output results in pages (-n to set items per page, default 10)',
            '--more'),
        enum=FlagArgument('Enumerate results', '--enumerate'),
        flavor_id=ValueArgument('filter by flavor id', ('--flavor-id')),
        image_id=ValueArgument('filter by image id', ('--image-id')),
        user_id=ValueArgument('filter by user id', ('--user-id')),
        user_name=ValueArgument('filter by user name', ('--user-name')),
        status=ValueArgument(
            'filter by status (ACTIVE, STOPPED, REBOOT, ERROR, etc.)',
            ('--status')),
        meta=KeyValueArgument('filter by metadata key=values', ('--metadata')),
        meta_like=KeyValueArgument(
            'print only if in key=value, the value is part of actual value',
            ('--metadata-like')),
    )

    def _add_user_name(self, servers):
        uuids = self._uuids2usernames(list(set(
                [srv['user_id'] for srv in servers] +
                [srv['tenant_id'] for srv in servers])))
        for srv in servers:
            srv['user_id'] += ' (%s)' % uuids[srv['user_id']]
            srv['tenant_id'] += ' (%s)' % uuids[srv['tenant_id']]
        return servers

    def _apply_common_filters(self, servers):
        common_filters = dict()
        if self['status']:
            common_filters['status'] = self['status']
        if self['user_id'] or self['user_name']:
            uuid = self['user_id'] or self._username2uuid(self['user_name'])
            common_filters['user_id'] = uuid
        return filter_dicts_by_dict(servers, common_filters)

    def _filter_by_image(self, servers):
        iid = self['image_id']
        return [srv for srv in servers if srv['image']['id'] == iid]

    def _filter_by_flavor(self, servers):
        fid = self['flavor_id']
        return [srv for srv in servers if (
            '%s' % srv['image']['id'] == '%s' % fid)]

    def _filter_by_metadata(self, servers):
        new_servers = []
        for srv in servers:
            if not 'metadata' in srv:
                continue
            meta = [dict(srv['metadata'])]
            if self['meta']:
                meta = filter_dicts_by_dict(meta, self['meta'])
            if meta and self['meta_like']:
                meta = filter_dicts_by_dict(
                    meta, self['meta_like'], exact_match=False)
            if meta:
                new_servers.append(srv)
        return new_servers

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.date
    def _run(self):
        withimage = bool(self['image_id'])
        withflavor = bool(self['flavor_id'])
        withmeta = bool(self['meta'] or self['meta_like'])
        withcommons = bool(
            self['status'] or self['user_id'] or self['user_name'])
        detail = self['detail'] or (
            withimage or withflavor or withmeta or withcommons)
        servers = self.client.list_servers(detail, self['since'])

        servers = self._filter_by_name(servers)
        servers = self._filter_by_id(servers)
        servers = self._apply_common_filters(servers)
        if withimage:
            servers = self._filter_by_image(servers)
        if withflavor:
            servers = self._filter_by_flavor(servers)
        if withmeta:
            servers = self._filter_by_metadata(servers)

        if self['detail'] and not (
                self['json_output'] or self['output_format']):
            servers = self._add_user_name(servers)
        elif not (self['detail'] or (
                self['json_output'] or self['output_format'])):
            remove_from_items(servers, 'links')
        if detail and not self['detail']:
            for srv in servers:
                for key in set(srv).difference(self.PERMANENTS):
                    srv.pop(key)
        kwargs = dict(with_enumeration=self['enum'])
        if self['more']:
            kwargs['out'] = StringIO()
            kwargs['title'] = ()
        if self['limit']:
            servers = servers[:self['limit']]
        self._print(servers, **kwargs)
        if self['more']:
            pager(kwargs['out'].getvalue())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(server_cmds)
class server_info(_init_cyclades, _optional_json):
    """Detailed information on a Virtual Machine
    Contains:
    - name, id, status, create/update dates
    - network interfaces
    - metadata (e.g., os, superuser) and diagnostics
    - hardware flavor and os image ids
    """

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id):
        vm = self.client.get_server_details(server_id)
        uuids = self._uuids2usernames([vm['user_id'], vm['tenant_id']])
        vm['user_id'] += ' (%s)' % uuids[vm['user_id']]
        vm['tenant_id'] += ' (%s)' % uuids[vm['tenant_id']]
        self._print(vm, self.print_dict)

    def main(self, server_id):
        super(self.__class__, self)._run()
        self._run(server_id=server_id)


class PersonalityArgument(KeyValueArgument):

    terms = (
        ('local-path', 'contents'),
        ('server-path', 'path'),
        ('owner', 'owner'),
        ('group', 'group'),
        ('mode', 'mode'))

    @property
    def value(self):
        return getattr(self, '_value', [])

    @value.setter
    def value(self, newvalue):
        if newvalue == self.default:
            return self.value
        self._value, input_dict = [], {}
        for i, terms in enumerate(newvalue):
            termlist = terms.split(',')
            if len(termlist) > len(self.terms):
                msg = 'Wrong number of terms (1<=terms<=%s)' % len(self.terms)
                raiseCLIError(CLISyntaxError(msg), details=howto_personality)

            for k, v in self.terms:
                prefix = '%s=' % k
                for item in termlist:
                    if item.lower().startswith(prefix):
                        input_dict[k] = item[len(k) + 1:]
                        break
                    item = None
                if item:
                    termlist.remove(item)

            try:
                path = input_dict['local-path']
            except KeyError:
                path = termlist.pop(0)
                if not path:
                    raise CLIInvalidArgument(
                        '--personality: No local path specified',
                        details=howto_personality)

            if not exists(path):
                raise CLIInvalidArgument(
                    '--personality: File %s does not exist' % path,
                    details=howto_personality)

            self._value.append(dict(path=path))
            with open(path) as f:
                self._value[i]['contents'] = b64encode(f.read())
            for k, v in self.terms[1:]:
                try:
                    self._value[i][v] = input_dict[k]
                except KeyError:
                    try:
                        self._value[i][v] = termlist.pop(0)
                    except IndexError:
                        continue


@command(server_cmds)
class server_create(_init_cyclades, _optional_json, _server_wait):
    """Create a server (aka Virtual Machine)
    Parameters:
    - name: (single quoted text)
    - flavor id: Hardware flavor. Pick one from: /flavor list
    - image id: OS images. Pick one from: /image list
    """

    arguments = dict(
        personality=PersonalityArgument(
            (80 * ' ').join(howto_personality), ('-p', '--personality')),
        wait=FlagArgument('Wait server to build', ('-w', '--wait')),
        cluster_size=IntArgument(
            'Create a cluster of servers of this size. In this case, the name'
            'parameter is the prefix of each server in the cluster (e.g.,'
            'srv1, srv2, etc.',
            '--cluster-size')
    )

    @errors.cyclades.cluster_size
    def _create_cluster(self, prefix, flavor_id, image_id, size):
        servers = [dict(
            name='%s%s' % (prefix, i if size > 1 else ''),
            flavor_id=flavor_id,
            image_id=image_id,
            personality=self['personality']) for i in range(1, 1 + size)]
        if size == 1:
            return [self.client.create_server(**servers[0])]
        try:
            r = self.client.async_run(self.client.create_server, servers)
            return r
        except Exception as e:
            if size == 1:
                raise e
            try:
                requested_names = [s['name'] for s in servers]
                spawned_servers = [dict(
                    name=s['name'],
                    id=s['id']) for s in self.client.list_servers() if (
                        s['name'] in requested_names)]
                self.error('Failed to build %s servers' % size)
                self.error('Found %s matching servers:' % len(spawned_servers))
                self._print(spawned_servers, out=self._err)
                self.error('Check if any of these servers should be removed\n')
            except Exception as ne:
                self.error('Error (%s) while notifying about errors' % ne)
            finally:
                raise e

    @errors.generic.all
    @errors.cyclades.connection
    @errors.plankton.id
    @errors.cyclades.flavor_id
    def _run(self, name, flavor_id, image_id):
        for r in self._create_cluster(
                name, flavor_id, image_id, size=self['cluster_size'] or 1):
            if not r:
                self.error('Create %s: server response was %s' % (name, r))
                continue
            usernames = self._uuids2usernames(
                [r['user_id'], r['tenant_id']])
            r['user_id'] += ' (%s)' % usernames[r['user_id']]
            r['tenant_id'] += ' (%s)' % usernames[r['tenant_id']]
            self._print(r, self.print_dict)
            if self['wait']:
                self._wait(r['id'], r['status'])
            self.writeln(' ')

    def main(self, name, flavor_id, image_id):
        super(self.__class__, self)._run()
        self._run(name=name, flavor_id=flavor_id, image_id=image_id)


@command(server_cmds)
class server_rename(_init_cyclades, _optional_output_cmd):
    """Set/update a virtual server name
    virtual server names are not unique, therefore multiple servers may share
    the same name
    """

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id, new_name):
        self._optional_output(
            self.client.update_server_name(int(server_id), new_name))

    def main(self, server_id, new_name):
        super(self.__class__, self)._run()
        self._run(server_id=server_id, new_name=new_name)


@command(server_cmds)
class server_delete(_init_cyclades, _optional_output_cmd, _server_wait):
    """Delete a virtual server"""

    arguments = dict(
        wait=FlagArgument('Wait server to be destroyed', ('-w', '--wait')),
        cluster=FlagArgument(
            '(DANGEROUS) Delete all virtual servers prefixed with the cluster '
            'prefix. In that case, the prefix replaces the server id',
            '--cluster')
    )

    def _server_ids(self, server_var):
        if self['cluster']:
            return [s['id'] for s in self.client.list_servers() if (
                s['name'].startswith(server_var))]

        @errors.cyclades.server_id
        def _check_server_id(self, server_id):
            return server_id

        return [_check_server_id(self, server_id=server_var), ]

    @errors.generic.all
    @errors.cyclades.connection
    def _run(self, server_var):
        for server_id in self._server_ids(server_var):
            if self['wait']:
                details = self.client.get_server_details(server_id)
                status = details['status']

            r = self.client.delete_server(server_id)
            self._optional_output(r)

            if self['wait']:
                self._wait(server_id, status)

    def main(self, server_id_or_cluster_prefix):
        super(self.__class__, self)._run()
        self._run(server_id_or_cluster_prefix)


@command(server_cmds)
class server_reboot(_init_cyclades, _optional_output_cmd, _server_wait):
    """Reboot a virtual server"""

    arguments = dict(
        hard=FlagArgument(
            'perform a hard reboot (deprecated)', ('-f', '--force')),
        type=ValueArgument('SOFT or HARD - default: SOFT', ('--type')),
        wait=FlagArgument('Wait server to be destroyed', ('-w', '--wait'))
    )

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id):
        hard_reboot = self['hard']
        if hard_reboot:
            self.error(
                'WARNING: -f/--force will be deprecated in version 0.12\n'
                '\tIn the future, please use --type=hard instead')
        if self['type']:
            if self['type'].lower() in ('soft', ):
                hard_reboot = False
            elif self['type'].lower() in ('hard', ):
                hard_reboot = True
            else:
                raise CLISyntaxError(
                    'Invalid reboot type %s' % self['type'],
                    importance=2, details=[
                        '--type values are either SOFT (default) or HARD'])

        r = self.client.reboot_server(int(server_id), hard_reboot)
        self._optional_output(r)

        if self['wait']:
            self._wait(server_id, 'REBOOT')

    def main(self, server_id):
        super(self.__class__, self)._run()
        self._run(server_id=server_id)


@command(server_cmds)
class server_start(_init_cyclades, _optional_output_cmd, _server_wait):
    """Start an existing virtual server"""

    arguments = dict(
        wait=FlagArgument('Wait server to be destroyed', ('-w', '--wait'))
    )

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id):
        status = 'ACTIVE'
        if self['wait']:
            details = self.client.get_server_details(server_id)
            status = details['status']
            if status in ('ACTIVE', ):
                return

        r = self.client.start_server(int(server_id))
        self._optional_output(r)

        if self['wait']:
            self._wait(server_id, status)

    def main(self, server_id):
        super(self.__class__, self)._run()
        self._run(server_id=server_id)


@command(server_cmds)
class server_shutdown(_init_cyclades, _optional_output_cmd, _server_wait):
    """Shutdown an active virtual server"""

    arguments = dict(
        wait=FlagArgument('Wait server to be destroyed', ('-w', '--wait'))
    )

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id):
        status = 'STOPPED'
        if self['wait']:
            details = self.client.get_server_details(server_id)
            status = details['status']
            if status in ('STOPPED', ):
                return

        r = self.client.shutdown_server(int(server_id))
        self._optional_output(r)

        if self['wait']:
            self._wait(server_id, status)

    def main(self, server_id):
        super(self.__class__, self)._run()
        self._run(server_id=server_id)


@command(server_cmds)
class server_console(_init_cyclades, _optional_json):
    """Get a VNC console to access an existing virtual server
    Console connection information provided (at least):
    - host: (url or address) a VNC host
    - port: (int) the gateway to enter virtual server on host
    - password: for VNC authorization
    """

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id):
        self._print(
            self.client.get_server_console(int(server_id)), self.print_dict)

    def main(self, server_id):
        super(self.__class__, self)._run()
        self._run(server_id=server_id)


@command(server_cmds)
class server_resize(_init_cyclades, _optional_output_cmd):
    """Set a different flavor for an existing server
    To get server ids and flavor ids:
    /server list
    /flavor list
    """

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    @errors.cyclades.flavor_id
    def _run(self, server_id, flavor_id):
        self._optional_output(self.client.resize_server(server_id, flavor_id))

    def main(self, server_id, flavor_id):
        super(self.__class__, self)._run()
        self._run(server_id=server_id, flavor_id=flavor_id)


@command(server_cmds)
class server_firewall(_init_cyclades):
    """Manage virtual server firewall profiles for public networks"""


@command(server_cmds)
class server_firewall_set(
        _init_cyclades, _optional_output_cmd, _firewall_wait):
    """Set the firewall profile on virtual server public network
    Values for profile:
    - DISABLED: Shutdown firewall
    - ENABLED: Firewall in normal mode
    - PROTECTED: Firewall in secure mode
    """

    arguments = dict(
        wait=FlagArgument('Wait server firewall to build', ('-w', '--wait')),
        timeout=IntArgument(
            'Set wait timeout in seconds (default: 60)', '--timeout',
            default=60)
    )

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    @errors.cyclades.firewall
    def _run(self, server_id, profile):
        if self['timeout'] and not self['wait']:
            raise CLIInvalidArgument('Invalid use of --timeout', details=[
                'Timeout is used only along with -w/--wait'])
        old_profile = self.client.get_firewall_profile(server_id)
        if old_profile.lower() == profile.lower():
            self.error('Firewall of server %s: allready in status %s' % (
                server_id, old_profile))
        else:
            self._optional_output(self.client.set_firewall_profile(
                server_id=int(server_id), profile=('%s' % profile).upper()))
            if self['wait']:
                self._wait(server_id, old_profile, timeout=self['timeout'])

    def main(self, server_id, profile):
        super(self.__class__, self)._run()
        self._run(server_id=server_id, profile=profile)


@command(server_cmds)
class server_firewall_get(_init_cyclades):
    """Get the firewall profile for a virtual servers' public network"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id):
        self.writeln(self.client.get_firewall_profile(server_id))

    def main(self, server_id):
        super(self.__class__, self)._run()
        self._run(server_id=server_id)


@command(server_cmds)
class server_addr(_init_cyclades, _optional_json):
    """List the addresses of all network interfaces on a virtual server"""

    arguments = dict(
        enum=FlagArgument('Enumerate results', '--enumerate')
    )

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id):
        reply = self.client.list_server_nics(int(server_id))
        self._print(reply, with_enumeration=self['enum'] and (reply) > 1)

    def main(self, server_id):
        super(self.__class__, self)._run()
        self._run(server_id=server_id)


@command(server_cmds)
class server_metadata(_init_cyclades):
    """Manage Server metadata (key:value pairs of server attributes)"""


@command(server_cmds)
class server_metadata_list(_init_cyclades, _optional_json):
    """Get server metadata"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    @errors.cyclades.metadata
    def _run(self, server_id, key=''):
        self._print(
            self.client.get_server_metadata(int(server_id), key),
            self.print_dict)

    def main(self, server_id, key=''):
        super(self.__class__, self)._run()
        self._run(server_id=server_id, key=key)


@command(server_cmds)
class server_metadata_set(_init_cyclades, _optional_json):
    """Set / update virtual server metadata
    Metadata should be given in key/value pairs in key=value format
    For example: /server metadata set <server id> key1=value1 key2=value2
    Old, unreferenced metadata will remain intact
    """

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id, keyvals):
        assert keyvals, 'Please, add some metadata ( key=value)'
        metadata = dict()
        for keyval in keyvals:
            k, sep, v = keyval.partition('=')
            if sep and k:
                metadata[k] = v
            else:
                raiseCLIError(
                    'Invalid piece of metadata %s' % keyval,
                    importance=2, details=[
                        'Correct metadata format: key=val',
                        'For example:',
                        '/server metadata set <server id>'
                        'key1=value1 key2=value2'])
        self._print(
            self.client.update_server_metadata(int(server_id), **metadata),
            self.print_dict)

    def main(self, server_id, *key_equals_val):
        super(self.__class__, self)._run()
        self._run(server_id=server_id, keyvals=key_equals_val)


@command(server_cmds)
class server_metadata_delete(_init_cyclades, _optional_output_cmd):
    """Delete virtual server metadata"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    @errors.cyclades.metadata
    def _run(self, server_id, key):
        self._optional_output(
            self.client.delete_server_metadata(int(server_id), key))

    def main(self, server_id, key):
        super(self.__class__, self)._run()
        self._run(server_id=server_id, key=key)


@command(server_cmds)
class server_stats(_init_cyclades, _optional_json):
    """Get virtual server statistics"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id):
        self._print(
            self.client.get_server_stats(int(server_id)), self.print_dict)

    def main(self, server_id):
        super(self.__class__, self)._run()
        self._run(server_id=server_id)


@command(server_cmds)
class server_wait(_init_cyclades, _server_wait):
    """Wait for server to finish [BUILD, STOPPED, REBOOT, ACTIVE]"""

    arguments = dict(
        timeout=IntArgument(
            'Wait limit in seconds (default: 60)', '--timeout', default=60)
    )

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id, current_status):
        r = self.client.get_server_details(server_id)
        if r['status'].lower() == current_status.lower():
            self._wait(server_id, current_status, timeout=self['timeout'])
        else:
            self.error(
                'Server %s: Cannot wait for status %s, '
                'status is already %s' % (
                    server_id, current_status, r['status']))

    def main(self, server_id, current_status='BUILD'):
        super(self.__class__, self)._run()
        self._run(server_id=server_id, current_status=current_status)


@command(flavor_cmds)
class flavor_list(_init_cyclades, _optional_json, _name_filter, _id_filter):
    """List available hardware flavors"""

    PERMANENTS = ('id', 'name')

    arguments = dict(
        detail=FlagArgument('show detailed output', ('-l', '--details')),
        limit=IntArgument('limit # of listed flavors', ('-n', '--number')),
        more=FlagArgument(
            'output results in pages (-n to set items per page, default 10)',
            '--more'),
        enum=FlagArgument('Enumerate results', '--enumerate'),
        ram=ValueArgument('filter by ram', ('--ram')),
        vcpus=ValueArgument('filter by number of VCPUs', ('--vcpus')),
        disk=ValueArgument('filter by disk size in GB', ('--disk')),
        disk_template=ValueArgument(
            'filter by disk_templace', ('--disk-template'))
    )

    def _apply_common_filters(self, flavors):
        common_filters = dict()
        if self['ram']:
            common_filters['ram'] = self['ram']
        if self['vcpus']:
            common_filters['vcpus'] = self['vcpus']
        if self['disk']:
            common_filters['disk'] = self['disk']
        if self['disk_template']:
            common_filters['SNF:disk_template'] = self['disk_template']
        return filter_dicts_by_dict(flavors, common_filters)

    @errors.generic.all
    @errors.cyclades.connection
    def _run(self):
        withcommons = self['ram'] or self['vcpus'] or (
            self['disk'] or self['disk_template'])
        detail = self['detail'] or withcommons
        flavors = self.client.list_flavors(detail)
        flavors = self._filter_by_name(flavors)
        flavors = self._filter_by_id(flavors)
        if withcommons:
            flavors = self._apply_common_filters(flavors)
        if not (self['detail'] or (
                self['json_output'] or self['output_format'])):
            remove_from_items(flavors, 'links')
        if detail and not self['detail']:
            for flv in flavors:
                for key in set(flv).difference(self.PERMANENTS):
                    flv.pop(key)
        kwargs = dict(out=StringIO(), title=()) if self['more'] else {}
        self._print(
            flavors,
            with_redundancy=self['detail'], with_enumeration=self['enum'],
            **kwargs)
        if self['more']:
            pager(kwargs['out'].getvalue())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(flavor_cmds)
class flavor_info(_init_cyclades, _optional_json):
    """Detailed information on a hardware flavor
    To get a list of available flavors and flavor ids, try /flavor list
    """

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.flavor_id
    def _run(self, flavor_id):
        self._print(
            self.client.get_flavor_details(int(flavor_id)), self.print_dict)

    def main(self, flavor_id):
        super(self.__class__, self)._run()
        self._run(flavor_id=flavor_id)


def _add_name(self, net):
        user_id, tenant_id, uuids = net['user_id'], net['tenant_id'], []
        if user_id:
            uuids.append(user_id)
        if tenant_id:
            uuids.append(tenant_id)
        if uuids:
            usernames = self._uuids2usernames(uuids)
            if user_id:
                net['user_id'] += ' (%s)' % usernames[user_id]
            if tenant_id:
                net['tenant_id'] += ' (%s)' % usernames[tenant_id]


@command(network_cmds)
class network_info(_init_cyclades, _optional_json):
    """Detailed information on a network
    To get a list of available networks and network ids, try /network list
    """

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.network_id
    def _run(self, network_id):
        network = self.client.get_network_details(int(network_id))
        _add_name(self, network)
        self._print(network, self.print_dict, exclude=('id'))

    def main(self, network_id):
        super(self.__class__, self)._run()
        self._run(network_id=network_id)


@command(network_cmds)
class network_list(_init_cyclades, _optional_json, _name_filter, _id_filter):
    """List networks"""

    PERMANENTS = ('id', 'name')

    arguments = dict(
        detail=FlagArgument('show detailed output', ('-l', '--details')),
        limit=IntArgument('limit # of listed networks', ('-n', '--number')),
        more=FlagArgument(
            'output results in pages (-n to set items per page, default 10)',
            '--more'),
        enum=FlagArgument('Enumerate results', '--enumerate'),
        status=ValueArgument('filter by status', ('--status')),
        public=FlagArgument('only public networks', ('--public')),
        private=FlagArgument('only private networks', ('--private')),
        dhcp=FlagArgument('show networks with dhcp', ('--with-dhcp')),
        no_dhcp=FlagArgument('show networks without dhcp', ('--without-dhcp')),
        user_id=ValueArgument('filter by user id', ('--user-id')),
        user_name=ValueArgument('filter by user name', ('--user-name')),
        gateway=ValueArgument('filter by gateway (IPv4)', ('--gateway')),
        gateway6=ValueArgument('filter by gateway (IPv6)', ('--gateway6')),
        cidr=ValueArgument('filter by cidr (IPv4)', ('--cidr')),
        cidr6=ValueArgument('filter by cidr (IPv6)', ('--cidr6')),
        type=ValueArgument('filter by type', ('--type')),
    )

    def _apply_common_filters(self, networks):
        common_filter = dict()
        if self['public']:
            if self['private']:
                return []
            common_filter['public'] = self['public']
        elif self['private']:
            common_filter['public'] = False
        if self['dhcp']:
            if self['no_dhcp']:
                return []
            common_filter['dhcp'] = True
        elif self['no_dhcp']:
            common_filter['dhcp'] = False
        if self['user_id'] or self['user_name']:
            uuid = self['user_id'] or self._username2uuid(self['user_name'])
            common_filter['user_id'] = uuid
        for term in ('status', 'gateway', 'gateway6', 'cidr', 'cidr6', 'type'):
            if self[term]:
                common_filter[term] = self[term]
        return filter_dicts_by_dict(networks, common_filter)

    def _add_name(self, networks, key='user_id'):
        uuids = self._uuids2usernames(
            list(set([net[key] for net in networks])))
        for net in networks:
            v = net.get(key, None)
            if v:
                net[key] += ' (%s)' % uuids[v]
        return networks

    @errors.generic.all
    @errors.cyclades.connection
    def _run(self):
        withcommons = False
        for term in (
                'status', 'public', 'private', 'user_id', 'user_name', 'type',
                'gateway', 'gateway6', 'cidr', 'cidr6', 'dhcp', 'no_dhcp'):
            if self[term]:
                withcommons = True
                break
        detail = self['detail'] or withcommons
        networks = self.client.list_networks(detail)
        networks = self._filter_by_name(networks)
        networks = self._filter_by_id(networks)
        if withcommons:
            networks = self._apply_common_filters(networks)
        if not (self['detail'] or (
                self['json_output'] or self['output_format'])):
            remove_from_items(networks, 'links')
        if detail and not self['detail']:
            for net in networks:
                for key in set(net).difference(self.PERMANENTS):
                    net.pop(key)
        if self['detail'] and not (
                self['json_output'] or self['output_format']):
            self._add_name(networks)
            self._add_name(networks, 'tenant_id')
        kwargs = dict(with_enumeration=self['enum'])
        if self['more']:
            kwargs['out'] = StringIO()
            kwargs['title'] = ()
        if self['limit']:
            networks = networks[:self['limit']]
        self._print(networks, **kwargs)
        if self['more']:
            pager(kwargs['out'].getvalue())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(network_cmds)
class network_create(_init_cyclades, _optional_json, _network_wait):
    """Create an (unconnected) network"""

    arguments = dict(
        cidr=ValueArgument('explicitly set cidr', '--with-cidr'),
        gateway=ValueArgument('explicitly set gateway', '--with-gateway'),
        dhcp=FlagArgument('Use dhcp (default: off)', '--with-dhcp'),
        type=ValueArgument(
            'Valid network types are '
            'CUSTOM, IP_LESS_ROUTED, MAC_FILTERED (default), PHYSICAL_VLAN',
            '--with-type',
            default='MAC_FILTERED'),
        wait=FlagArgument('Wait network to build', ('-w', '--wait'))
    )

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.network_max
    def _run(self, name):
        r = self.client.create_network(
            name,
            cidr=self['cidr'],
            gateway=self['gateway'],
            dhcp=self['dhcp'],
            type=self['type'])
        _add_name(self, r)
        self._print(r, self.print_dict)
        if self['wait'] and r['status'] in ('PENDING', ):
            self._wait(r['id'], 'PENDING')

    def main(self, name):
        super(self.__class__, self)._run()
        self._run(name)


@command(network_cmds)
class network_rename(_init_cyclades, _optional_output_cmd):
    """Set the name of a network"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.network_id
    def _run(self, network_id, new_name):
        self._optional_output(
                self.client.update_network_name(int(network_id), new_name))

    def main(self, network_id, new_name):
        super(self.__class__, self)._run()
        self._run(network_id=network_id, new_name=new_name)


@command(network_cmds)
class network_delete(_init_cyclades, _optional_output_cmd, _network_wait):
    """Delete a network"""

    arguments = dict(
        wait=FlagArgument('Wait network to build', ('-w', '--wait'))
    )

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.network_in_use
    @errors.cyclades.network_id
    def _run(self, network_id):
        status = 'DELETED'
        if self['wait']:
            r = self.client.get_network_details(network_id)
            status = r['status']
            if status in ('DELETED', ):
                return

        r = self.client.delete_network(int(network_id))
        self._optional_output(r)

        if self['wait']:
            self._wait(network_id, status)

    def main(self, network_id):
        super(self.__class__, self)._run()
        self._run(network_id=network_id)


@command(network_cmds)
class network_connect(_init_cyclades, _optional_output_cmd):
    """Connect a server to a network"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    @errors.cyclades.network_id
    def _run(self, server_id, network_id):
        self._optional_output(
                self.client.connect_server(int(server_id), int(network_id)))

    def main(self, server_id, network_id):
        super(self.__class__, self)._run()
        self._run(server_id=server_id, network_id=network_id)


@command(network_cmds)
class network_disconnect(_init_cyclades):
    """Disconnect a nic that connects a server to a network
    Nic ids are listed as "attachments" in detailed network information
    To get detailed network information: /network info <network id>
    """

    @errors.cyclades.nic_format
    def _server_id_from_nic(self, nic_id):
        return nic_id.split('-')[1]

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    @errors.cyclades.nic_id
    def _run(self, nic_id, server_id):
        num_of_disconnected = self.client.disconnect_server(server_id, nic_id)
        if not num_of_disconnected:
            raise ClientError(
                'Network Interface %s not found on server %s' % (
                    nic_id, server_id),
                status=404)
        print('Disconnected %s connections' % num_of_disconnected)

    def main(self, nic_id):
        super(self.__class__, self)._run()
        server_id = self._server_id_from_nic(nic_id=nic_id)
        self._run(nic_id=nic_id, server_id=server_id)


@command(network_cmds)
class network_wait(_init_cyclades, _network_wait):
    """Wait for server to finish [PENDING, ACTIVE, DELETED]"""

    arguments = dict(
        timeout=IntArgument(
            'Wait limit in seconds (default: 60)', '--timeout', default=60)
    )

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.network_id
    def _run(self, network_id, current_status):
        net = self.client.get_network_details(network_id)
        if net['status'].lower() == current_status.lower():
            self._wait(network_id, current_status, timeout=self['timeout'])
        else:
            self.error(
                'Network %s: Cannot wait for status %s, '
                'status is already %s' % (
                    network_id, current_status, net['status']))

    def main(self, network_id, current_status='PENDING'):
        super(self.__class__, self)._run()
        self._run(network_id=network_id, current_status=current_status)


@command(ip_cmds)
class ip_pools(_init_cyclades, _optional_json):
    """List pools of floating IPs"""

    @errors.generic.all
    @errors.cyclades.connection
    def _run(self):
        r = self.client.get_floating_ip_pools()
        self._print(r if self['json_output'] or self['output_format'] else r[
            'floating_ip_pools'])

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(ip_cmds)
class ip_list(_init_cyclades, _optional_json):
    """List reserved floating IPs"""

    @errors.generic.all
    @errors.cyclades.connection
    def _run(self):
        r = self.client.get_floating_ips()
        self._print(r if self['json_output'] or self['output_format'] else r[
            'floating_ips'])

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(ip_cmds)
class ip_info(_init_cyclades, _optional_json):
    """Details for an IP"""

    @errors.generic.all
    @errors.cyclades.connection
    def _run(self, ip):
        self._print(self.client.get_floating_ip(ip), self.print_dict)

    def main(self, IP):
        super(self.__class__, self)._run()
        self._run(ip=IP)


@command(ip_cmds)
class ip_reserve(_init_cyclades, _optional_json):
    """Reserve a floating IP
    An IP is reserved from an IP pool. The default IP pool is chosen
    automatically, but there is the option if specifying an explicit IP pool.
    """

    arguments = dict(pool=ValueArgument('Source IP pool', ('--pool'), None))

    @errors.generic.all
    @errors.cyclades.connection
    def _run(self, ip=None):
        self._print([self.client.alloc_floating_ip(self['pool'], ip)])

    def main(self, requested_IP=None):
        super(self.__class__, self)._run()
        self._run(ip=requested_IP)


@command(ip_cmds)
class ip_release(_init_cyclades, _optional_output_cmd):
    """Release a floating IP
    The release IP is "returned" to the IP pool it came from.
    """

    @errors.generic.all
    @errors.cyclades.connection
    def _run(self, ip):
        self._optional_output(self.client.delete_floating_ip(ip))

    def main(self, IP):
        super(self.__class__, self)._run()
        self._run(ip=IP)


@command(ip_cmds)
class ip_attach(_init_cyclades, _optional_output_cmd):
    """Attach a floating IP to a server
    """

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id, ip):
        self._optional_output(self.client.attach_floating_ip(server_id, ip))

    def main(self, server_id, IP):
        super(self.__class__, self)._run()
        self._run(server_id=server_id, ip=IP)


@command(ip_cmds)
class ip_detach(_init_cyclades, _optional_output_cmd):
    """Detach a floating IP from a server
    """

    @errors.generic.all
    @errors.cyclades.connection
    @errors.cyclades.server_id
    def _run(self, server_id, ip):
        self._optional_output(self.client.detach_floating_ip(server_id, ip))

    def main(self, server_id, IP):
        super(self.__class__, self)._run()
        self._run(server_id=server_id, ip=IP)



# Copyright 2012-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.command

from kamaki.cli.commands import errors
from kamaki.cli import command
from kamaki.cli.commands import _command_init
from kamaki.cli.command_tree import CommandTree
from kamaki.clients import livetest
from kamaki.cli.errors import raiseCLIError

livetest_cmds = CommandTree('livetest', 'Client func. tests on live servers')
_commands = [livetest_cmds]


class _livetest_init(_command_init):

    def _run(self, client, method=None):
        if method:
            livetest.main([client, method], config=self.config)
        else:
            livetest.main([client], config=self.config)

    def main(self, client, method=None):
        return self._run(client, method)


@command(livetest_cmds)
class livetest_error(_livetest_init):
    """Create an error message with optional message"""

    @errors.generic.all
    def _run(self, errmsg='', importance=0, index=0):
        l = [1, 2]
        try:
            l[int(index)]
        except Exception as err:
            raiseCLIError(err, errmsg, importance)
        raiseCLIError(None, errmsg, importance)

    def main(self, errmsg='', importance=0, index=0):
        self._run(errmsg, importance, index)


@command(livetest_cmds)
class livetest_args(_livetest_init):
    """Test how arguments are treated by kamaki"""

    @errors.generic.all
    def _run(self, *args):
        self.writeln(args)

    def main(self, *args):
        self._run(args)


@command(livetest_cmds)
class livetest_all(_livetest_init):
    """test all clients"""

    @errors.generic.all
    def _run(self):
        for client in ('pithos', 'cyclades', 'image', 'astakos'):
            super(self.__class__, self)._run(client)

    def main(self):
        self._run()


@command(livetest_cmds)
class livetest_pithos(_livetest_init):
    """ test Pithos client"""

    @errors.generic.all
    def _run(self, method=None):
        super(self.__class__, self)._run('pithos', method)

    def main(self, method=None):
        self._run(method)


@command(livetest_cmds)
class livetest_cyclades(_livetest_init):
    """ test Cyclades client"""

    @errors.generic.all
    def _run(self, method=None):
        super(self.__class__, self)._run('cyclades', method)

    def main(self, method=None):
        self._run(method)


@command(livetest_cmds)
class livetest_image(_livetest_init):
    """ test Image client"""

    @errors.generic.all
    def _run(self, method=None):
        super(self.__class__, self)._run('image', method)

    def main(self, method=None):
        self._run(method)


@command(livetest_cmds)
class livetest_astakos(_livetest_init):
    """ test Astakos client"""

    @errors.generic.all
    def _run(self, method=None):
        super(self.__class__, self)._run('astakos', method)

    def main(self, method=None):
        self._run(method)


@command(livetest_cmds)
class livetest_prints(_livetest_init):
    """ user-test print methods for lists and dicts"""

    d1 = {'key0a': 'val0a', 'key0b': 'val0b', 'key0c': 'val0c'}

    l1 = [1, 'string', '3', 'many (2 or 3) numbers and strings combined', 5]

    d2 = {'id': 'val0a', 'key0b': d1, 'title': l1}

    l2 = [d2, l1, d1]

    spr_msg = 'long key of size 75 characters is used to check the effects on'
    spr_msg += ' total result for long messages that drive pep8 completely mad'
    d3 = {'dict 1': d1, 'dict 2': d2, 'list2': l2, spr_msg: l1}

    @errors.generic.all
    def _run(self):
        from kamaki.cli.utils import print_dict, print_list, print_items
        self.writeln('Test simple dict:\n- - -')
        print_dict(self.d1)
        self.writeln('- - -\n')
        self.writeln('\nTest simple list:\n- - -')
        print_list(self.l1)
        self.writeln('- - -\n')
        self.writeln('\nTest 2-level dict:\n- - -')
        print_dict(self.d2)
        self.writeln('- - -\n')
        self.writeln('\nTest non-trivial list:\n- - -')
        print_list(self.l2)
        self.writeln('- - -')
        self.writeln('\nTest extreme dict:\n- - -')
        print_dict(self.d3)
        self.writeln('- - -\n')
        self.writeln('Test simple enumerated dict:\n- - -')
        print_dict(self.d1, with_enumeration=True)
        self.writeln('- - -\n')
        self.writeln('\nTest simple enumerated list:\n- - -')
        print_list(self.l1, with_enumeration=True)
        self.writeln('- - -\n')
        self.writeln('Test non-trivial deep-enumerated dict:\n- - -')
        print_dict(self.d2, with_enumeration=True, recursive_enumeration=True)
        self.writeln('- - -\n')
        self.writeln('\nTest non-trivial enumerated list:\n- - -')
        print_list(self.l2, with_enumeration=True)
        self.writeln('- - -\n')
        self.writeln('\nTest print_items with id:\n- - -')
        print_items([
            {'id': '42', 'title': 'lalakis 1', 'content': self.d1},
            {'id': '142', 'title': 'lalakis 2', 'content': self.d2}])
        self.writeln('- - -')
        self.writeln('\nTest print_items with id and enumeration:\n- - -')
        print_items(
            [
                {'id': '42', 'title': 'lalakis 1', 'content': self.d1},
                {'id': '142', 'title': 'lalakis 2', 'content': self.d2}],
            with_enumeration=True)
        self.writeln('- - -')
        self.writeln('\nTest print_items with id, title, redundancy:\n- - -')
        print_items(
            [
                {'id': '42', 'title': 'lalakis 1', 'content': self.d1},
                {'id': '142', 'title': 'lalakis 2', 'content': self.d2}],
            title=('id', 'title'),
            with_redundancy=True)
        self.writeln('- - -')
        self.writeln('\nTest print_items with lists- - -')
        print_items([['i00', 'i01', 'i02'], [self.l2, 'i11', self.d1], 3])
        self.writeln('- - -')

    def main(self):
        self._run()



# Copyright 2011-2012 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.


"""Classes to support pools of arbitrary objects.

The :class:`ObjectPool` class in this module abstracts a pool
of arbitrary objects. Subclasses need to define the details regarding
creation, destruction, allocation and release of their specific objects.

"""

# This should work under gevent, because gevent monkey patches 'threading'
# if not, we can detect if running under gevent, e.g. using
# if 'gevent' in sys.modules:
#     from gevent.coros import Semaphore
# else:
#     from threading import Semaphore
from threading import Semaphore, Lock


__all__ = ['ObjectPool', 'ObjectPoolError',
           'PoolLimitError', 'PoolVerificationError']

import logging
log = logging.getLogger(__name__)


class ObjectPoolError(Exception):
    pass


class PoolLimitError(ObjectPoolError):
    pass


class PoolVerificationError(ObjectPoolError):
    pass


class ObjectPool(object):
    """Generic Object Pool.

    The pool consists of an object set and an allocation semaphore.

    pool_get() gets an allocation from the semaphore
               and an object from the pool set.

    pool_put() releases an allocation to the semaphore
               and puts an object back to the pool set.

    Subclasses must implement these thread-safe hooks:
    _pool_create()
            used as a subclass hook to auto-create new objects in pool_get().
    _pool_verify()
            verifies objects before they are returned by pool_get()
    _pool_cleanup()
            cleans up and verifies objects before their return by pool_put().

    While allocations are strictly accounted for and limited by
    the semaphore, objects are expendable:

    The hook provider and the caller are solely responsible for object
    handling.

    pool_get() may create an object if there is none in the pool set.
    pool_get() may return no object, leaving object creation to the caller.
    pool_put() may return no object to the pool set.
    Objects to pool_put() to the pool need not be those from pool_get().
    Objects to pool_get() need not be those from pool_put().


    Callers beware:
    The pool limit size must be greater than the total working set of objects,
    otherwise it will hang. When in doubt, use an impossibly large size limit.
    Since the pool grows on demand, this will not waste resources.
    However, in that case, the pool must not be used as a flow control device
    (i.e. relying on pool_get() blocking to stop threads),
    as the impossibly large pool size limit will defer blocking until too late.

    """
    def __init__(self, size=None):
        try:
            self.size = int(size)
            assert size >= 1
        except:
            raise ValueError("Invalid size for pool (positive integer "
                             "required): %r" % (size,))

        self._semaphore = Semaphore(size)  # Pool grows up to size limit
        self._mutex = Lock()  # Protect shared _set oject
        self._set = set()
        log.debug("Initialized pool %r", self)

    def __repr__(self):
        return ("<pool %d: size=%d, len(_set)=%d, semaphore=%d>" %
                (id(self), self.size, len(self._set),
                 self._semaphore._Semaphore__value))

    def pool_get(self, blocking=True, timeout=None, create=True, verify=True):
        """Get an object from the pool.

        Get a pool allocation and an object from the pool set.
        Raise PoolLimitError if the pool allocation limit has been reached.
        If the pool set is empty, create a new object (create==True),
        or return None (create==False) and let the caller create it.
        All objects returned (except None) are verified.

        """
        # timeout argument only supported by gevent and py3k variants
        # of Semaphore. acquire() will raise TypeError if timeout
        # is specified but not supported by the underlying implementation.
        log.debug("GET: about to get object from pool %r", self)
        kw = {"blocking": blocking}
        if timeout is not None:
            kw["timeout"] = timeout
        sema = self._semaphore
        r = sema.acquire(**kw)
        if not r:
            raise PoolLimitError()

        try:
            created = 0
            while 1:
                with self._mutex:
                    try:
                        obj = self._set.pop()
                    except KeyError:
                        obj = None
                if obj is None and create:
                    obj = self._pool_create()
                    created = 1

                if not self._pool_verify(obj):
                    if created:
                        m = "Pool %r cannot verify new object %r" % (self, obj)
                        raise PoolVerificationError(m)
                    continue
                break
        except:
            sema.release()
            raise

        # We keep _semaphore acquired, put() will release it
        log.debug("GOT: object %r from pool %r", obj, self)
        return obj

    def pool_put(self, obj):
        """Put an object back into the pool.

        Release an allocation and return an object to the pool.
        If obj is None, or _pool_cleanup returns True,
        then the allocation is released,
        but no object returned to the pool set

        """
        log.debug("PUT-BEFORE: about to put object %r back to pool %r",
                  obj, self)
        if obj is not None and not self._pool_cleanup(obj):
            with self._mutex:
                if obj in self._set:
                    log.warning("Object %r already in _set of pool %r",
                                obj, self)
                self._set.add(obj)
        self._semaphore.release()
        log.debug("PUT-AFTER: finished putting object %r back to pool %r",
                  obj, self)

    def pool_create_free(self):
        """Create a free new object that is not put into the pool.

        Just for convenience, let the users create objects with
        the exact same configuration as those that are used with the pool

        """
        obj = self._pool_create_free()
        return obj

    def _pool_create_free(self):
        """Create a free new object that is not put into the pool.

        This should be overriden by pool classes.
        Otherwise, it just calls _pool_create().

        """
        return self._pool_create()

    def _pool_create(self):
        """Create a new object to be used with this pool.

        Create a new object to be used with this pool,
        should be overriden in subclasses.
        Must be thread-safe.

        """
        raise NotImplementedError

    def _pool_verify(self, obj):
        """Verify an object after getting it from the pool.

        If it returns False, the object is discarded
        and another one is drawn from the pool.
        If the pool is empty, a new object is created.
        If the new object fails to verify, pool_get() will fail.
        Must be thread-safe.

        """
        raise NotImplementedError

    def _pool_cleanup(self, obj):
        """Cleanup an object before being put back into the pool.

        Cleanup an object before it can be put back into the pull,
        ensure it is in a stable, reusable state.
        Must be thread-safe.

        """
        raise NotImplementedError


class PooledObject(object):
    """Generic Object Pool context manager and pooled object proxy.

    The PooledObject instance acts as a context manager to
    be used in a with statement:

        with PooledObject(...) as obj:
            use(obj)

    The with block above is roughly equivalent to:

        pooled = PooledObject(...):
        try:
            obj = pooled.acquire()
            assert(obj is pooled.obj)
            use(obj)
        finally:
            pooled.release()

    After exiting the with block, or releasing,
    the code MUST not use the obj again in any way.

    """

    # NOTE: We need all definitions at class-level
    # to avoid infinite __gettatr__() recursion.
    # This is also true for subclasses.

    # NOTE: Typically you will only need to override
    #       __init__() and get_pool

    # Initialization. Do not customize.
    _pool_settings = None
    _pool_get_settings = None
    _pool_kwargs = None
    _pool = None
    obj = None

    #####################################################
    ### Subclass attribute customization begins here. ###

    _pool_log_prefix = "POOL"
    _pool_class = ObjectPool

    # default keyword args to pass to pool initialization
    _pool_default_settings = (
        ('size', 25),
    )

    # keyword args to pass to pool_get
    _pool_default_get_settings = (
        ('blocking', True),
        #('timeout', None),
        ('create', True),
        ('verify', True),
    )

    # behavior settings
    _pool_attach_context = False
    _pool_disable_after_release = True
    _pool_ignore_double_release = False

    ###  Subclass attribute customization ends here.  ###
    #####################################################

    def __init__(self, pool_settings=None, get_settings=None,
                 attach_context=None, disable_after_release=None,
                 ignore_double_release=None, **kwargs):
        """Initialize a PooledObject instance.

        Accept only keyword arguments.
        Some of them are filtered for this instance's configuration,
        and the rest are saved in ._pool_kwargs for later use.

        The filtered keywords are:

        pool_settings:  keyword args forwarded to pool instance initialization
                        in get_pool(), on top of the class defaults.
                        If not given, the remaining keyword args are
                        forwarded instead.

        get_settings:   keyword args forwarded to the pool's .pool_get() on top
                        of the class defaults.

        attach_context: boolean overriding the class default.
                        If True, after getting an object from the pool,
                        attach self onto it before returning it,
                        so that the context manager caller can have
                        access to the manager object within the with: block.

        disable_after_release:
                        boolean overriding the class default.
                        If True, the PooledObject will not allow a second
                        acquisition after the first release. For example,
                        the second with will raise an AssertionError:
                        manager = PooledObject()
                        with manager as c:
                            pass
                        with manager as c:
                            pass

        ignore_double_release:
                        boolean overriding the class default.
                        If True, the PooledObject will allow consecutive
                        calls to release the underlying pooled object.
                        Only the first one has an effect.
                        If False, an AssertionError is raised.

        """
        self._pool_kwargs = kwargs
        self._pool = None
        self.obj = None

        _get_settings = dict(self._pool_default_get_settings)
        if get_settings is not None:
            _get_settings.update(get_settings)
        self._pool_get_settings = _get_settings

        if attach_context is not None:
            self._pool_attach_context = attach_context

        if pool_settings is None:
            pool_settings = kwargs

        _pool_settings = dict(self._pool_default_settings)
        _pool_settings.update(**pool_settings)
        self._pool_settings = _pool_settings

    def get_pool(self):
        """Return a suitable pool object to work with.

        Called within .acquire(), it is meant to be
        overriden by sublasses, to create a new pool,
        or retrieve an existing one, based on the PooledObject
        initialization keywords stored in self._pool_kwargs.

        """
        pool = self._pool_class(**self._pool_settings)
        return pool

    ### Maybe overriding get_pool() and __init__() above is enough ###

    def __repr__(self):
        return ("<object %s of class %s: "
                "proxy for object (%r) in pool (%r)>" % (
                id(self), self.__class__.__name__,
                self.obj, self._pool))

    __str__ = __repr__

    ## Proxy the real object. Disabled until needed.
    ##
    ##def __getattr__(self, name):
    ##    return getattr(self.obj, name)

    ##def __setattr__(self, name, value):
    ##    if hasattr(self, name):
    ##        _setattr = super(PooledObject, self).__setattr__
    ##        _setattr(name, value)
    ##    else:
    ##        setattr(self.obj, name, value)

    ##def __delattr_(self, name):
    ##    _delattr = super(PooledObject, self).__delattr__
    ##    if hasattr(self, name):
    ##        _delattr(self, name)
    ##    else:
    ##        delattr(self.obj, name)

    def __enter__(self):
        return self.acquire()

    def __exit__(self, exc_type, exc_value, trace):
        return self.release()

    def acquire(self):
        log.debug("%s Acquiring (context: %r)", self._pool_log_prefix, self)
        pool = self._pool
        if pool is False:
            m = "%r: has been released. No further pool access allowed." % (
                self,)
            raise AssertionError(m)
        if pool is not None:
            m = "Double acquire in %r" % self
            raise AssertionError(m)

        pool = self.get_pool()
        self._pool = pool

        obj = pool.pool_get(**self._pool_get_settings)
        if self._pool_attach_context:
            obj._pool_context = self
        self.obj = obj
        log.debug("%s Acquired %r", self._pool_log_prefix, obj)
        return obj

    def release(self):
        log.debug("%s Releasing (context: %r)", self._pool_log_prefix, self)
        pool = self._pool
        if pool is None:
            m = "%r: no pool" % (self,)
            raise AssertionError(m)

        obj = self.obj
        if obj is None:
            if self._pool_ignore_double_release:
                return
            m = "%r: no object. Double release?" % (self,)
            raise AssertionError(m)

        pool.pool_put(obj)
        self.obj = None
        if self._pool_disable_after_release:
            self._pool = False



# Copyright 2012-2013 GRNET S.A. All rights reserved.
#
# Redistribution and use in source and binary forms, with or
# without modification, are permitted provided that the following
# conditions are met:
#
#   1. Redistributions of source code must retain the above
#      copyright notice, this list of conditions and the following
#      disclaimer.
#
#   2. Redistributions in binary form must reproduce the above
#      copyright notice, this list of conditions and the following
#      disclaimer in the documentation and/or other materials
#      provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY GRNET S.A. ``AS IS'' AND ANY EXPRESS
# OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
# PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL GRNET S.A OR
# CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
# USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
# AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.
#
# The views and conclusions contained in the software and
# documentation are those of the authors and should not be
# interpreted as representing official policies, either expressed
# or implied, of GRNET S.A.command

from json import load, dumps
from os import path
from logging import getLogger
from io import StringIO
from pydoc import pager

from kamaki.cli import command
from kamaki.cli.command_tree import CommandTree
from kamaki.cli.utils import filter_dicts_by_dict
from kamaki.clients.image import ImageClient
from kamaki.clients.pithos import PithosClient
from kamaki.clients.astakos import AstakosClient
from kamaki.clients import ClientError
from kamaki.cli.argument import (
    FlagArgument, ValueArgument, RepeatableArgument, KeyValueArgument,
    IntArgument, ProgressBarArgument)
from kamaki.cli.commands.cyclades import _init_cyclades
from kamaki.cli.errors import raiseCLIError, CLIBaseUrlError
from kamaki.cli.commands import _command_init, errors, addLogSettings
from kamaki.cli.commands import (
    _optional_output_cmd, _optional_json, _name_filter, _id_filter)


image_cmds = CommandTree(
    'image',
    'Cyclades/Plankton API image commands\n'
    'image compute:\tCyclades/Compute API image commands')
_commands = [image_cmds]


howto_image_file = [
    'Kamaki commands to:',
    ' get current user id: /user authenticate',
    ' check available containers: /file list',
    ' create a new container: /file create <container>',
    ' check container contents: /file list <container>',
    ' upload files: /file upload <image file> <container>',
    ' register an image: /image register <image name> <container>:<path>']

about_image_id = ['To see a list of available image ids: /image list']


log = getLogger(__name__)


class _init_image(_command_init):
    @errors.generic.all
    @addLogSettings
    def _run(self):
        if getattr(self, 'cloud', None):
            img_url = self._custom_url('image') or self._custom_url('plankton')
            if img_url:
                token = self._custom_token('image') or self._custom_token(
                    'plankton') or self.config.get_cloud(self.cloud, 'token')
                self.client = ImageClient(base_url=img_url, token=token)
                return
        if getattr(self, 'auth_base', False):
            plankton_endpoints = self.auth_base.get_service_endpoints(
                self._custom_type('image') or self._custom_type(
                    'plankton') or 'image',
                self._custom_version('image') or self._custom_version(
                    'plankton') or '')
            base_url = plankton_endpoints['publicURL']
            token = self.auth_base.token
        else:
            raise CLIBaseUrlError(service='plankton')
        self.client = ImageClient(base_url=base_url, token=token)

    def main(self):
        self._run()


# Plankton Image Commands


def _validate_image_meta(json_dict, return_str=False):
    """
    :param json_dict" (dict) json-formated, of the form
        {"key1": "val1", "key2": "val2", ...}

    :param return_str: (boolean) if true, return a json dump

    :returns: (dict) if return_str is not True, else return str

    :raises TypeError, AttributeError: Invalid json format

    :raises AssertionError: Valid json but invalid image properties dict
    """
    json_str = dumps(json_dict, indent=2)
    for k, v in json_dict.items():
        if k.lower() == 'properties':
            for pk, pv in v.items():
                prop_ok = not (isinstance(pv, dict) or isinstance(pv, list))
                assert prop_ok, 'Invalid property value for key %s' % pk
                key_ok = not (' ' in k or '-' in k)
                assert key_ok, 'Invalid property key %s' % k
            continue
        meta_ok = not (isinstance(v, dict) or isinstance(v, list))
        assert meta_ok, 'Invalid value for meta key %s' % k
        meta_ok = ' ' not in k
        assert meta_ok, 'Invalid meta key [%s]' % k
        json_dict[k] = '%s' % v
    return json_str if return_str else json_dict


def _load_image_meta(filepath):
    """
    :param filepath: (str) the (relative) path of the metafile

    :returns: (dict) json_formated

    :raises TypeError, AttributeError: Invalid json format

    :raises AssertionError: Valid json but invalid image properties dict
    """
    with open(path.abspath(filepath)) as f:
        meta_dict = load(f)
        try:
            return _validate_image_meta(meta_dict)
        except AssertionError:
            log.debug('Failed to load properties from file %s' % filepath)
            raise


def _validate_image_location(location):
    """
    :param location: (str) pithos://<user-id>/<container>/<image-path>

    :returns: (<user-id>, <container>, <image-path>)

    :raises AssertionError: if location is invalid
    """
    prefix = 'pithos://'
    msg = 'Invalid prefix for location %s , try: %s' % (location, prefix)
    assert location.startswith(prefix), msg
    service, sep, rest = location.partition('://')
    assert sep and rest, 'Location %s is missing user-id' % location
    uuid, sep, rest = rest.partition('/')
    assert sep and rest, 'Location %s is missing container' % location
    container, sep, img_path = rest.partition('/')
    assert sep and img_path, 'Location %s is missing image path' % location
    return uuid, container, img_path


@command(image_cmds)
class image_list(_init_image, _optional_json, _name_filter, _id_filter):
    """List images accessible by user"""

    PERMANENTS = (
        'id', 'name',
        'status', 'container_format', 'disk_format', 'size')

    arguments = dict(
        detail=FlagArgument('show detailed output', ('-l', '--details')),
        container_format=ValueArgument(
            'filter by container format',
            '--container-format'),
        disk_format=ValueArgument('filter by disk format', '--disk-format'),
        size_min=IntArgument('filter by minimum size', '--size-min'),
        size_max=IntArgument('filter by maximum size', '--size-max'),
        status=ValueArgument('filter by status', '--status'),
        owner=ValueArgument('filter by owner', '--owner'),
        owner_name=ValueArgument('filter by owners username', '--owner-name'),
        order=ValueArgument(
            'order by FIELD ( - to reverse order)',
            '--order',
            default=''),
        limit=IntArgument('limit number of listed images', ('-n', '--number')),
        more=FlagArgument(
            'output results in pages (-n to set items per page, default 10)',
            '--more'),
        enum=FlagArgument('Enumerate results', '--enumerate'),
        prop=KeyValueArgument('filter by property key=value', ('--property')),
        prop_like=KeyValueArgument(
            'fliter by property key=value where value is part of actual value',
            ('--property-like')),
    )

    def _filter_by_owner(self, images):
        ouuid = self['owner'] or self._username2uuid(self['owner_name'])
        return filter_dicts_by_dict(images, dict(owner=ouuid))

    def _add_owner_name(self, images):
        uuids = self._uuids2usernames(
            list(set([img['owner'] for img in images])))
        for img in images:
            img['owner'] += ' (%s)' % uuids[img['owner']]
        return images

    def _filter_by_properties(self, images):
        new_images = []
        for img in images:
            props = [dict(img['properties'])]
            if self['prop']:
                props = filter_dicts_by_dict(props, self['prop'])
            if props and self['prop_like']:
                props = filter_dicts_by_dict(
                    props, self['prop_like'], exact_match=False)
            if props:
                new_images.append(img)
        return new_images

    @errors.generic.all
    @errors.cyclades.connection
    def _run(self):
        super(self.__class__, self)._run()
        filters = {}
        for arg in set([
                'container_format',
                'disk_format',
                'name',
                'size_min',
                'size_max',
                'status']).intersection(self.arguments):
            filters[arg] = self[arg]

        order = self['order']
        detail = self['detail'] or (
            self['prop'] or self['prop_like']) or (
            self['owner'] or self['owner_name'])

        images = self.client.list_public(detail, filters, order)

        if self['owner'] or self['owner_name']:
            images = self._filter_by_owner(images)
        if self['prop'] or self['prop_like']:
            images = self._filter_by_properties(images)
        images = self._filter_by_id(images)
        images = self._non_exact_name_filter(images)

        if self['detail'] and not (
                self['json_output'] or self['output_format']):
            images = self._add_owner_name(images)
        elif detail and not self['detail']:
            for img in images:
                for key in set(img).difference(self.PERMANENTS):
                    img.pop(key)
        kwargs = dict(with_enumeration=self['enum'])
        if self['limit']:
            images = images[:self['limit']]
        if self['more']:
            kwargs['out'] = StringIO()
            kwargs['title'] = ()
        self._print(images, **kwargs)
        if self['more']:
            pager(kwargs['out'].getvalue())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(image_cmds)
class image_meta(_init_image):
    """Manage image metadata and custom properties"""


@command(image_cmds)
class image_info(_init_image, _optional_json):
    """Get image metadata
    Image metadata include:
    - image file information (location, size, etc.)
    - image information (id, name, etc.)
    - image os properties (os, fs, etc.)
    """

    @errors.generic.all
    @errors.plankton.connection
    @errors.plankton.id
    def _run(self, image_id):
        meta = self.client.get_meta(image_id)
        if not (self['json_output'] or self['output_format']):
            meta['owner'] += ' (%s)' % self._uuid2username(meta['owner'])
        self._print(meta, self.print_dict)

    def main(self, image_id):
        super(self.__class__, self)._run()
        self._run(image_id=image_id)


@command(image_cmds)
class image_meta_set(_init_image, _optional_output_cmd):
    """Add / update metadata and properties for an image
    The original image preserves the values that are not affected
    """

    arguments = dict(
        name=ValueArgument('Set a new name', ('--name')),
        disk_format=ValueArgument('Set a new disk format', ('--disk-format')),
        container_format=ValueArgument(
            'Set a new container format', ('--container-format')),
        status=ValueArgument('Set a new status', ('--status')),
        publish=FlagArgument('publish the image', ('--publish')),
        unpublish=FlagArgument('unpublish the image', ('--unpublish')),
        properties=KeyValueArgument(
            'set property in key=value form (can be repeated)',
            ('-p', '--property'))
    )

    def _check_empty(self):
        for term in (
                'name', 'disk_format', 'container_format', 'status', 'publish',
                'unpublish', 'properties'):
            if self['term']:
                if self['publish'] and self['unpublish']:
                    raiseCLIError(
                        '--publish and --unpublish are mutually exclusive')
                return
        raiseCLIError(
            'Nothing to update, please use arguments (-h for a list)')

    @errors.generic.all
    @errors.plankton.connection
    @errors.plankton.id
    def _run(self, image_id):
        self._check_empty()
        meta = self.client.get_meta(image_id)
        for k, v in self['properties'].items():
            meta['properties'][k.upper()] = v
        self._optional_output(self.client.update_image(
            image_id,
            name=self['name'],
            disk_format=self['disk_format'],
            container_format=self['container_format'],
            status=self['status'],
            public=self['publish'] or self['unpublish'] or None,
            **meta['properties']))

    def main(self, image_id):
        super(self.__class__, self)._run()
        self._run(image_id=image_id)


@command(image_cmds)
class image_meta_delete(_init_image, _optional_output_cmd):
    """Remove/empty image metadata and/or custom properties"""

    arguments = dict(
        disk_format=FlagArgument('Empty disk format', ('--disk-format')),
        container_format=FlagArgument(
            'Empty container format', ('--container-format')),
        status=FlagArgument('Empty status', ('--status')),
        properties=RepeatableArgument(
            'Property keys to remove', ('-p', '--property'))
    )

    def _check_empty(self):
        for t in ('disk_format', 'container_format', 'status', 'properties'):
            if self[t]:
                return
        raiseCLIError(
            'Nothing to update, please use arguments (-h for a list)')

    @errors.generic.all
    @errors.plankton.connection
    @errors.plankton.id
    def _run(self, image_id):
        self._check_empty()
        meta = self.client.get_meta(image_id)
        for k in self['properties']:
            meta['properties'].pop(k.upper(), None)
        self._optional_output(self.client.update_image(
            image_id,
            disk_format='' if self['disk_format'] else None,
            container_format='' if self['container_format'] else None,
            status='' if self['status'] else None,
            **meta['properties']))

    def main(self, image_id):
        super(self.__class__, self)._run()
        self._run(image_id=image_id)


@command(image_cmds)
class image_register(_init_image, _optional_json):
    """(Re)Register an image file to an Image service
    The image file must be stored at a pithos repository
    Some metadata can be set by user (e.g., disk-format) while others are set
    only automatically (e.g., image id). There are also some custom user
    metadata, called properties.
    A register command creates a remote meta file at
    .  <container>:<image path>.meta
    Users may download and edit this file and use it to re-register one or more
    images.
    In case of a meta file, runtime arguments for metadata or properties
    override meta file settings.
    """

    container_info_cache = {}

    arguments = dict(
        checksum=ValueArgument('Set image checksum', '--checksum'),
        container_format=ValueArgument(
            'Set container format', '--container-format'),
        disk_format=ValueArgument('Set disk format', '--disk-format'),
        owner_name=ValueArgument('Set user uuid by user name', '--owner-name'),
        properties=KeyValueArgument(
            'Add property (user-specified metadata) in key=value form'
            '(can be repeated)',
            ('-p', '--property')),
        is_public=FlagArgument('Mark image as public', '--public'),
        size=IntArgument('Set image size in bytes', '--size'),
        metafile=ValueArgument(
            'Load metadata from a json-formated file <img-file>.meta :'
            '{"key1": "val1", "key2": "val2", ..., "properties: {...}"}',
            ('--metafile')),
        metafile_force=FlagArgument(
            'Overide remote metadata file', ('-f', '--force')),
        no_metafile_upload=FlagArgument(
            'Do not store metadata in remote meta file',
            ('--no-metafile-upload')),
        container=ValueArgument(
            'Pithos+ container containing the image file',
            ('-C', '--container')),
        uuid=ValueArgument('Custom user uuid', '--uuid'),
        local_image_path=ValueArgument(
            'Local image file path to upload and register '
            '(still need target file in the form container:remote-path )',
            '--upload-image-file'),
        progress_bar=ProgressBarArgument(
            'Do not use progress bar', '--no-progress-bar', default=False)
    )

    def _get_user_id(self):
        atoken = self.client.token
        if getattr(self, 'auth_base', False):
            return self.auth_base.term('id', atoken)
        else:
            astakos_url = self.config.get('user', 'url') or self.config.get(
                'astakos', 'url')
            if not astakos_url:
                raise CLIBaseUrlError(service='astakos')
            user = AstakosClient(astakos_url, atoken)
            return user.term('id')

    def _get_pithos_client(self, container):
        if self['no_metafile_upload']:
            return None
        ptoken = self.client.token
        if getattr(self, 'auth_base', False):
            pithos_endpoints = self.auth_base.get_service_endpoints(
                'object-store')
            purl = pithos_endpoints['publicURL']
        else:
            purl = self.config.get_cloud('pithos', 'url')
        if not purl:
            raise CLIBaseUrlError(service='pithos')
        return PithosClient(purl, ptoken, self._get_user_id(), container)

    def _store_remote_metafile(self, pclient, remote_path, metadata):
        return pclient.upload_from_string(
            remote_path, _validate_image_meta(metadata, return_str=True),
            container_info_cache=self.container_info_cache)

    def _load_params_from_file(self, location):
        params, properties = dict(), dict()
        pfile = self['metafile']
        if pfile:
            try:
                for k, v in _load_image_meta(pfile).items():
                    key = k.lower().replace('-', '_')
                    if key == 'properties':
                        for pk, pv in v.items():
                            properties[pk.upper().replace('-', '_')] = pv
                    elif key == 'name':
                            continue
                    elif key == 'location':
                        if location:
                            continue
                        location = v
                    else:
                        params[key] = v
            except Exception as e:
                raiseCLIError(e, 'Invalid json metadata config file')
        return params, properties, location

    def _load_params_from_args(self, params, properties):
        for key in set([
                'checksum',
                'container_format',
                'disk_format',
                'owner',
                'size',
                'is_public']).intersection(self.arguments):
            params[key] = self[key]
        for k, v in self['properties'].items():
            properties[k.upper().replace('-', '_')] = v

    def _validate_location(self, location):
        if not location:
            raiseCLIError(
                'No image file location provided',
                importance=2, details=[
                    'An image location is needed. Image location format:',
                    '  <container>:<path>',
                    ' where an image file at the above location must exist.'
                    ] + howto_image_file)
        try:
            return _validate_image_location(location)
        except AssertionError as ae:
            raiseCLIError(
                ae, 'Invalid image location format',
                importance=1, details=[
                    'Valid image location format:',
                    '  <container>:<img-file-path>'
                    ] + howto_image_file)

    @staticmethod
    def _old_location_format(location):
        prefix = 'pithos://'
        try:
            if location.startswith(prefix):
                uuid, sep, rest = location[len(prefix):].partition('/')
                container, sep, path = rest.partition('/')
                return (uuid, container, path)
        except Exception as e:
            raiseCLIError(e, 'Invalid location format', details=[
                'Correct location format:', '  <container>:<image path>'])
        return ()

    def _mine_location(self, container_path):
        old_response = self._old_location_format(container_path)
        if old_response:
            return old_response
        uuid = self['uuid'] or (self._username2uuid(self['owner_name']) if (
                    self['owner_name']) else self._get_user_id())
        if not uuid:
            if self['owner_name']:
                raiseCLIError('No user with username %s' % self['owner_name'])
            raiseCLIError('Failed to get user uuid', details=[
                'For details on current user:',
                '  /user whoami',
                'To authenticate a new user through a user token:',
                '  /user authenticate <token>'])
        if self['container']:
            return uuid, self['container'], container_path
        container, sep, path = container_path.partition(':')
        if not (bool(container) and bool(path)):
            raiseCLIError(
                'Incorrect container-path format', importance=1, details=[
                'Use : to seperate container form path',
                '  <container>:<image-path>',
                'OR',
                'Use -C to specifiy a container',
                '  -C <container> <image-path>'] + howto_image_file)

        return uuid, container, path

    @errors.generic.all
    @errors.plankton.connection
    @errors.pithos.container
    def _run(self, name, uuid, dst_cont, img_path):
        if self['local_image_path']:
            with open(self['local_image_path']) as f:
                pithos = self._get_pithos_client(dst_cont)
                (pbar, upload_cb) = self._safe_progress_bar('Uploading')
                if pbar:
                    hash_bar = pbar.clone()
                    hash_cb = hash_bar.get_generator('Calculating hashes')
                pithos.upload_object(
                    img_path, f,
                    hash_cb=hash_cb, upload_cb=upload_cb,
                    container_info_cache=self.container_info_cache)
                pbar.finish()

        location = 'pithos://%s/%s/%s' % (uuid, dst_cont, img_path)
        (params, properties, new_loc) = self._load_params_from_file(location)
        if location != new_loc:
            uuid, dst_cont, img_path = self._validate_location(new_loc)
        self._load_params_from_args(params, properties)
        pclient = self._get_pithos_client(dst_cont)

        #check if metafile exists
        meta_path = '%s.meta' % img_path
        if pclient and not self['metafile_force']:
            try:
                pclient.get_object_info(meta_path)
                raiseCLIError(
                    'Metadata file %s:%s already exists, abort' % (
                        dst_cont, meta_path),
                    details=['Registration ABORTED', 'Try -f to overwrite'])
            except ClientError as ce:
                if ce.status != 404:
                    raise

        #register the image
        try:
            r = self.client.register(name, location, params, properties)
        except ClientError as ce:
            if ce.status in (400, 404):
                raiseCLIError(
                    ce, 'Nonexistent image file location\n\t%s' % location,
                    details=[
                        'Does the image file %s exist at container %s ?' % (
                            img_path, dst_cont)] + howto_image_file)
            raise
        r['owner'] += ' (%s)' % self._uuid2username(r['owner'])
        self._print(r, self.print_dict)

        #upload the metadata file
        if pclient:
            try:
                meta_headers = pclient.upload_from_string(
                    meta_path, dumps(r, indent=2),
                    container_info_cache=self.container_info_cache)
            except TypeError:
                self.error(
                    'Failed to dump metafile %s:%s' % (dst_cont, meta_path))
                return
            if self['json_output'] or self['output_format']:
                self.print_json(dict(
                    metafile_location='%s:%s' % (dst_cont, meta_path),
                    headers=meta_headers))
            else:
                self.error('Metadata file uploaded as %s:%s (version %s)' % (
                    dst_cont, meta_path, meta_headers['x-object-version']))

    def main(self, name, container___image_path):
        super(self.__class__, self)._run()
        self._run(name, *self._mine_location(container___image_path))


@command(image_cmds)
class image_unregister(_init_image, _optional_output_cmd):
    """Unregister an image (does not delete the image file)"""

    @errors.generic.all
    @errors.plankton.connection
    @errors.plankton.id
    def _run(self, image_id):
        self._optional_output(self.client.unregister(image_id))

    def main(self, image_id):
        super(self.__class__, self)._run()
        self._run(image_id=image_id)


@command(image_cmds)
class image_shared(_init_image, _optional_json):
    """List images shared by a member"""

    @errors.generic.all
    @errors.plankton.connection
    def _run(self, member):
        r = self.client.list_shared(member)
        if r:
            uuid = self._username2uuid(member)
            r = self.client.list_shared(uuid) if uuid else []
        self._print(r, title=('image_id',))

    def main(self, member_id_or_username):
        super(self.__class__, self)._run()
        self._run(member_id_or_username)


@command(image_cmds)
class image_members(_init_image):
    """Manage members. Members of an image are users who can modify it"""


@command(image_cmds)
class image_members_list(_init_image, _optional_json):
    """List members of an image"""

    @errors.generic.all
    @errors.plankton.connection
    @errors.plankton.id
    def _run(self, image_id):
        members = self.client.list_members(image_id)
        if not (self['json_output'] or self['output_format']):
            uuids = [member['member_id'] for member in members]
            usernames = self._uuids2usernames(uuids)
            for member in members:
                member['member_id'] += ' (%s)' % usernames[member['member_id']]
        self._print(members, title=('member_id',))

    def main(self, image_id):
        super(self.__class__, self)._run()
        self._run(image_id=image_id)


@command(image_cmds)
class image_members_add(_init_image, _optional_output_cmd):
    """Add a member to an image"""

    @errors.generic.all
    @errors.plankton.connection
    @errors.plankton.id
    def _run(self, image_id=None, member=None):
            self._optional_output(self.client.add_member(image_id, member))

    def main(self, image_id, member_id):
        super(self.__class__, self)._run()
        self._run(image_id=image_id, member=member_id)


@command(image_cmds)
class image_members_delete(_init_image, _optional_output_cmd):
    """Remove a member from an image"""

    @errors.generic.all
    @errors.plankton.connection
    @errors.plankton.id
    def _run(self, image_id=None, member=None):
            self._optional_output(self.client.remove_member(image_id, member))

    def main(self, image_id, member):
        super(self.__class__, self)._run()
        self._run(image_id=image_id, member=member)


@command(image_cmds)
class image_members_set(_init_image, _optional_output_cmd):
    """Set the members of an image"""

    @errors.generic.all
    @errors.plankton.connection
    @errors.plankton.id
    def _run(self, image_id, members):
            self._optional_output(self.client.set_members(image_id, members))

    def main(self, image_id, *member_ids):
        super(self.__class__, self)._run()
        self._run(image_id=image_id, members=member_ids)

# Compute Image Commands


@command(image_cmds)
class image_compute(_init_cyclades):
    """Cyclades/Compute API image commands"""


@command(image_cmds)
class image_compute_list(
        _init_cyclades, _optional_json, _name_filter, _id_filter):
    """List images"""

    PERMANENTS = ('id', 'name')

    arguments = dict(
        detail=FlagArgument('show detailed output', ('-l', '--details')),
        limit=IntArgument('limit number listed images', ('-n', '--number')),
        more=FlagArgument('handle long lists of results', '--more'),
        enum=FlagArgument('Enumerate results', '--enumerate'),
        user_id=ValueArgument('filter by user_id', '--user-id'),
        user_name=ValueArgument('filter by username', '--user-name'),
        meta=KeyValueArgument(
            'filter by metadata key=value (can be repeated)', ('--metadata')),
        meta_like=KeyValueArgument(
            'filter by metadata key=value (can be repeated)',
            ('--metadata-like'))
    )

    def _filter_by_metadata(self, images):
        new_images = []
        for img in images:
            meta = [dict(img['metadata'])]
            if self['meta']:
                meta = filter_dicts_by_dict(meta, self['meta'])
            if meta and self['meta_like']:
                meta = filter_dicts_by_dict(
                    meta, self['meta_like'], exact_match=False)
            if meta:
                new_images.append(img)
        return new_images

    def _filter_by_user(self, images):
        uuid = self['user_id'] or self._username2uuid(self['user_name'])
        return filter_dicts_by_dict(images, dict(user_id=uuid))

    def _add_name(self, images, key='user_id'):
        uuids = self._uuids2usernames(
            list(set([img[key] for img in images])))
        for img in images:
            img[key] += ' (%s)' % uuids[img[key]]
        return images

    @errors.generic.all
    @errors.cyclades.connection
    def _run(self):
        withmeta = bool(self['meta'] or self['meta_like'])
        withuser = bool(self['user_id'] or self['user_name'])
        detail = self['detail'] or withmeta or withuser
        images = self.client.list_images(detail)
        images = self._filter_by_name(images)
        images = self._filter_by_id(images)
        if withuser:
            images = self._filter_by_user(images)
        if withmeta:
            images = self._filter_by_metadata(images)
        if self['detail'] and not (
                self['json_output'] or self['output_format']):
            images = self._add_name(self._add_name(images, 'tenant_id'))
        elif detail and not self['detail']:
            for img in images:
                for key in set(img).difference(self.PERMANENTS):
                    img.pop(key)
        kwargs = dict(with_enumeration=self['enum'])
        if self['limit']:
            images = images[:self['limit']]
        if self['more']:
            kwargs['out'] = StringIO()
            kwargs['title'] = ()
        self._print(images, **kwargs)
        if self['more']:
            pager(kwargs['out'].getvalue())

    def main(self):
        super(self.__class__, self)._run()
        self._run()


@command(image_cmds)
class image_compute_info(_init_cyclades, _optional_json):
    """Get detailed information on an image"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.plankton.id
    def _run(self, image_id):
        image = self.client.get_image_details(image_id)
        uuids = [image['user_id'], image['tenant_id']]
        usernames = self._uuids2usernames(uuids)
        image['user_id'] += ' (%s)' % usernames[image['user_id']]
        image['tenant_id'] += ' (%s)' % usernames[image['tenant_id']]
        self._print(image, self.print_dict)

    def main(self, image_id):
        super(self.__class__, self)._run()
        self._run(image_id=image_id)


@command(image_cmds)
class image_compute_delete(_init_cyclades, _optional_output_cmd):
    """Delete an image (WARNING: image file is also removed)"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.plankton.id
    def _run(self, image_id):
        self._optional_output(self.client.delete_image(image_id))

    def main(self, image_id):
        super(self.__class__, self)._run()
        self._run(image_id=image_id)


@command(image_cmds)
class image_compute_properties(_init_cyclades):
    """Manage properties related to OS installation in an image"""


@command(image_cmds)
class image_compute_properties_list(_init_cyclades, _optional_json):
    """List all image properties"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.plankton.id
    def _run(self, image_id):
        self._print(self.client.get_image_metadata(image_id), self.print_dict)

    def main(self, image_id):
        super(self.__class__, self)._run()
        self._run(image_id=image_id)


@command(image_cmds)
class image_compute_properties_get(_init_cyclades, _optional_json):
    """Get an image property"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.plankton.id
    @errors.plankton.metadata
    def _run(self, image_id, key):
        self._print(
            self.client.get_image_metadata(image_id, key), self.print_dict)

    def main(self, image_id, key):
        super(self.__class__, self)._run()
        self._run(image_id=image_id, key=key)


@command(image_cmds)
class image_compute_properties_set(_init_cyclades, _optional_json):
    """Add / update a set of properties for an image
    properties must be given in the form key=value, e.v.
    /image compute properties set <image-id> key1=val1 key2=val2
    """

    @errors.generic.all
    @errors.cyclades.connection
    @errors.plankton.id
    def _run(self, image_id, keyvals):
        meta = dict()
        for keyval in keyvals:
            key, sep, val = keyval.partition('=')
            meta[key] = val
        self._print(
            self.client.update_image_metadata(image_id, **meta),
            self.print_dict)

    def main(self, image_id, *key_equals_value):
        super(self.__class__, self)._run()
        self._run(image_id=image_id, keyvals=key_equals_value)


@command(image_cmds)
class image_compute_properties_delete(_init_cyclades, _optional_output_cmd):
    """Delete a property from an image"""

    @errors.generic.all
    @errors.cyclades.connection
    @errors.plankton.id
    @errors.plankton.metadata
    def _run(self, image_id, key):
        self._optional_output(self.client.delete_image_metadata(image_id, key))

    def main(self, image_id, key):
        super(self.__class__, self)._run()
        self._run(image_id=image_id, key=key)
